# -*- coding: utf-8 -*-
"""API model for working with system configuration."""
from ...exceptions import ApiError, NotFoundError
from ..mixins import ChildMixins
from ..parsers.config import (config_build, config_unchanged, config_unknown,
                              parse_settings)
from ..parsers.tables import tablize


class SettingsChild(ChildMixins):
    """Child API object to work with system settings."""

    TITLE = ""

    @property
    def router_path(self):
        """Pass."""
        raise NotImplementedError  # pragma: no cover

    def get(self):
        """Get the current system settings.

        Returns:
            :obj:`dict`: current system settings
        """
        return parse_settings(raw=self._get(), title=self.TITLE)

    def get_section(self, section, sub_section=None, full_config=False):
        """Pass."""
        settings = self.get()
        title = settings["settings_title"]

        valid_sections = []

        for name, meta in settings["sections"].items():
            valid_sections.append(
                {
                    "Section Name": name,
                    "Section Title": meta["title"],
                    "Sub Section Names": "\n".join(list(meta["sub_sections"])),
                }
            )

            if name == section:
                if full_config:
                    meta["full_config"] = settings["config"]
                return meta

        err = f"Section Name {section!r} not found in {title}"
        raise NotFoundError(tablize(value=valid_sections, err=err))

    def get_sub_section(self, section, sub_section, full_config=False):
        """Pass."""
        settings = self.get_section(section=section, full_config=full_config)
        title = settings["settings_title"]

        if not settings["sub_sections"]:
            raise ApiError(f"Section Name {section!r} has no sub sections!")

        valids = []

        for name, meta in settings["sub_sections"].items():
            valids.append(
                {
                    "Sub Section Name": meta["name"],
                    "Sub Section Title": meta["title"],
                    "Section Name": meta["parent_name"],
                    "Section Title": meta["parent_title"],
                }
            )

            if name == sub_section:
                if full_config:
                    meta["full_config"] = settings["full_config"]
                return meta

        err = (
            f"Sub Section Name {sub_section!r} not found in under "
            f"Section Name {section!r} in {title}"
        )
        raise NotFoundError(tablize(value=valids, err=err))

    def update_section(self, section, **kwargs):
        """Update the system settings."""
        settings = self.get_section(section=section, full_config=True)
        title = settings["settings_title"]
        schemas = settings["schemas"]
        source = f"{title} Section Name {section!r}"
        old_config = settings["config"]
        full_config = settings["full_config"]

        new_config = {}
        new_config.update(kwargs)

        config_unknown(
            schemas=schemas, new_config=new_config, source=source,
        )
        config_build(
            schemas=schemas, old_config=old_config, new_config=new_config, source=source,
        )
        config_unchanged(
            schemas=schemas, old_config=old_config, new_config=new_config, source=source,
        )

        full_config[section] = new_config

        self._update(new_config=full_config)

        return self.get_section(section=section)

    def update_sub_section(self, section, sub_section, **kwargs):
        """Update the system settings."""
        settings = self.get_sub_section(
            section=section, sub_section=sub_section, full_config=True
        )
        title = settings["settings_title"]
        schemas = settings["schemas"]
        source = f"{title} Section Name {section!r} Sub Section Name {sub_section!r}"
        old_config = settings["config"]
        full_config = settings["full_config"]

        new_config = {}
        new_config.update(kwargs)

        config_unknown(
            schemas=schemas, new_config=new_config, source=source,
        )
        config_build(
            schemas=schemas, old_config=old_config, new_config=new_config, source=source,
        )
        config_unchanged(
            schemas=schemas, old_config=old_config, new_config=new_config, source=source,
        )

        full_config[section][sub_section] = new_config
        self._update(new_config=full_config)

        return self.get_sub_section(section=section, sub_section=sub_section)

    def _get(self):
        """Direct API method to get the current system settings.

        Returns:
            :obj:`dict`: current system settings
        """
        return self.request(method="get", path=self.router_path)

    def _update(self, new_config):
        """Direct API method to update the system settings."""
        return self.request(method="post", path=self.router_path, json=new_config)


class SettingsCore(SettingsChild):
    """Child API object to work with System Global Settings."""

    TITLE = "Global Settings"

    @property
    def router_path(self):
        """Route path for this setting object."""
        return self.parent.router.settings_core


class SettingsLifecycle(SettingsChild):
    """Child API object to work with Lifecycle Global Settings."""

    TITLE = "Lifecycle Settings"

    @property
    def router_path(self):
        """Route path for this setting object."""
        return self.parent.router.settings_lifecycle


class SettingsGui(SettingsChild):
    """Child API object to work with GUI Global Settings."""

    TITLE = "GUI Settings"

    @property
    def router_path(self):
        """Route path for this setting object."""
        return self.parent.router.settings_gui



# -*- coding: utf-8 -*-
"""Test suite."""
import pytest


class TestSystemNodes:
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_system):
        """Pass."""
        return api_system.nodes

    def test_get(self, apiobj):
        """Pass."""
        data = apiobj.get()
        assert isinstance(data, list)



# -*- coding: utf-8 -*-
"""Easy all-in-one connection handler."""
import re

import requests

from .api.adapters import Adapters
from .api.assets import Devices, Users
from .api.enforcements import Enforcements
from .api.system import System
from .auth import ApiKey
from .constants import (LOG_FILE_MAX_FILES, LOG_FILE_MAX_MB, LOG_FILE_NAME,
                        LOG_FILE_PATH, LOG_LEVEL_API, LOG_LEVEL_AUTH,
                        LOG_LEVEL_CONSOLE, LOG_LEVEL_FILE, LOG_LEVEL_HTTP,
                        LOG_LEVEL_PACKAGE, TIMEOUT_CONNECT, TIMEOUT_RESPONSE)
from .exceptions import ConnectError, InvalidCredentials
from .http import Http
from .logs import LOG, add_file, add_stderr, set_log_level


class Connect:
    """Easy all-in-one connection handler."""

    REASON_RES = [
        re.compile(r".*?object at.*?\>\: ([a-zA-Z0-9\]\[: ]+)"),
        re.compile(r".*?\] (.*) "),
    ]
    """:obj:`list` of patterns`: patterns to look for in exceptions that we can
    pretty up for user display."""

    def __init__(
        self,
        url,
        key,
        secret,
        wraperror=True,
        timeout_connect=TIMEOUT_CONNECT,
        timeout_response=TIMEOUT_RESPONSE,
        certpath=None,
        certverify=False,
        certwarn=True,
        cert_client_key=None,
        cert_client_cert=None,
        cert_client_both=None,
        proxy=None,
        save_history=False,
        log_request_attrs=None,
        log_response_attrs=None,
        log_request_body=False,
        log_response_body=False,
        log_logger=LOG,
        log_level_package=LOG_LEVEL_PACKAGE,
        log_level_http=LOG_LEVEL_HTTP,
        log_level_auth=LOG_LEVEL_AUTH,
        log_level_api=LOG_LEVEL_API,
        log_level_console=LOG_LEVEL_CONSOLE,
        log_level_file=LOG_LEVEL_FILE,
        log_console=False,
        log_file=False,
        log_file_name=LOG_FILE_NAME,
        log_file_path=LOG_FILE_PATH,
        log_file_max_mb=LOG_FILE_MAX_MB,
        log_file_max_files=LOG_FILE_MAX_FILES,
    ):
        """Easy all-in-one connection handler.

        Args:
            url (:obj:`str`): URL, hostname, or IP address of Axonius instance
            key (:obj:`str`): API Key from account page in Axonius instance
            secret (:obj:`str`): API Secret from account page in Axonius instance
            timeout_connect (:obj:`int`, optional):
                default :data:`TIMEOUT_CONNECT` - seconds to
                wait for connections to open to :attr:`url`
            timeout_response (:obj:`int`, optional):
                default :data:`TIMEOUT_RESPONSE` - seconds to
                wait for responses from :attr:`url`
            wraperror (:obj:`bool`, optional): default ``True``

                * if ``True`` wrap exceptions so that they are more user friendly
                * if ``False`` print the original exception with the full traceback
            certpath (:obj:`str`, optional): default ``None`` -
                path to CA bundle file to use when verifing certs offered by :attr:`url`
                instead of the system CA bundle
            certverify (:obj:`bool`, optional): default ``False`` - control
                validation of certs offered by :attr:`url`:

                * if ``True`` raise exception if cert is invalid/self-signed
                * if ``False`` only raise exception if cert is invalid
            certwarn (:obj:`bool`, optional): default ``True`` - show warnings from
                requests about certs offered by :attr:`url` that are self signed:

                * if ``True`` show warning only the first time it happens
                * if ``False`` never show warning
                * if ``None`` show warning every time it happens
            cert_client_key (:obj:`str`, optional):
                default ``None`` - path to private key file for cert_client_cert to
                offer to :attr:`url` (*must also supply cert_client_cert*)
            cert_client_cert (:obj:`str`, optional):
                default ``None`` - path to cert file to offer to :attr:`url`
                (*must also supply cert_client_key*)
            cert_client_both (:obj:`str`, optional):
                default ``None`` - path to cert file containing both the private key and
                cert to offer to :attr:`url`
            proxy (:obj:`str`, optional): default ``None`` - proxy to use
                when making https requests to :attr:`url`
            save_history (:obj:`bool`, optional): default ``True`` -

                * if ``True`` append responses to
                  :attr:`axonius_api_client.http.Http.HISTORY`
                * if ``False`` do not append responses to
                  :attr:`axonius_api_client.http.Http.HISTORY`
            log_request_attrs (:obj:`bool`): default ``None`` - control logging
              of request attributes:

              * if ``True``, log request attributes defined in
                :data:`axonius_api_client.LOG_REQUEST_ATTRS_VERBOSE`
              * if ``False``, log request attributes defined in
                :data:`axonius_api_client.LOG_REQUEST_ATTRS_BRIEF`
              * if ``None``, do not log any request attributes
            log_response_attrs (:obj:`bool`): default ``None`` - control logging
              of response attributes:

              * if ``True``, log response attributes defined in
                :data:`axonius_api_client.LOG_RESPONSE_ATTRS_VERBOSE`
              * if ``False``, log response attributes defined in
                :data:`axonius_api_client.LOG_RESPONSE_ATTRS_BRIEF`
              * if ``None``, do not log any response attributes
            log_request_body (:obj:`bool`): default ``False`` - control logging
              of request bodies:

              * if ``True``, log request bodies
              * if ``False``, do not log request bodies
            log_response_body (:obj:`bool`): default ``False`` - control logging
              of response bodies:

              * if ``True``, log response bodies
              * if ``False``, do not log response bodies

            log_logger (:obj:`logging.Logger`, optional):
                default :data:`axonius_api_client.LOG`
                logger to use as package root logger
            log_level_package (:obj:`str`, optional):
                default :data:`axonius_api_client.LOG_LEVEL_PACKAGE`
                log level to use for log_logger
            log_level_http (:obj:`str`, optional):
                default :data:`axonius_api_client.LOG_LEVEL_HTTP`
                log level to use for :obj:`axonius_api_client.http.Http`
            log_level_auth (:obj:`str`, optional):
                default :data:`axonius_api_client.LOG_LEVEL_AUTH`
                log level to use for all subclasses of
                :obj:`axonius_api_client.auth.Mixins`
            log_level_api (:obj:`str`, optional):
                default :data:`axonius_api_client.LOG_LEVEL_API`
                log level to use for all subclasses of
                :obj:`axonius_api_client.mixins.Mixins`
            log_level_console (:obj:`str`, optional):
                default :data:`axonius_api_client.LOG_LEVEL_CONSOLE`
                log level to use for logs sent to console
            log_level_file (:obj:`str`, optional):
                default :data:`axonius_api_client.LOG_LEVEL_FILE`
                log level to use for logs sent to file
            log_console (:obj:`bool`, optional): default ``False`` -

                * if ``True``, enable logging to console
                * if ``False``, do not log to console
            log_file (:obj:`bool`, optional): default ``False`` -

                * if ``True``, enable logging to file
                * if ``False``, do not log to console
            log_file_name (:obj:`str`, optional):
                default :data:`axonius_api_client.LOG_FILE_NAME`
                name of file to write logs to under log_file_path
            log_file_path (:obj:`str`, optional):
                default :data:`axonius_api_client.LOG_FILE_PATH`
                path to write log_file_name to
            log_file_max_mb (:obj:`str`, optional):
                default :data:`axonius_api_client.LOG_FILE_MAX_MB`
                rollover file logs at this many MB
            log_file_max_files (:obj:`str`, optional):
                default :data:`axonius_api_client.LOG_FILE_MAX_FILES`
                number of rollover file logs to keep
        """
        self._started = False
        self._wraperror = wraperror
        self.url = url
        """:obj:`str`: URL to connect to"""

        set_log_level(obj=log_logger, level=log_level_package)

        self._handler_file = None
        self._handler_con = None

        if log_console:
            self._handler_con = add_stderr(obj=log_logger, level=log_level_console)

        if log_file:
            self._handler_file = add_file(
                obj=log_logger,
                level=log_level_file,
                file_path=log_file_path,
                file_name=log_file_name,
                max_mb=log_file_max_mb,
                max_files=log_file_max_files,
            )

        self._http_args = {
            "url": url,
            "https_proxy": proxy,
            "certpath": certpath,
            "certwarn": certwarn,
            "certverify": certverify,
            "cert_client_both": cert_client_both,
            "cert_client_cert": cert_client_cert,
            "cert_client_key": cert_client_key,
            "log_level": log_level_http,
            "log_request_attrs": log_request_attrs,
            "log_response_attrs": log_response_attrs,
            "log_request_body": log_request_body,
            "log_response_body": log_response_body,
            "save_history": save_history,
            "connect_timeout": timeout_connect,
            "response_timeout": timeout_response,
        }

        self._auth_args = {"key": key, "secret": secret, "log_level": log_level_auth}

        self._http = Http(**self._http_args)

        self._auth = ApiKey(http=self._http, **self._auth_args)

        self._api_args = {"auth": self._auth, "log_level": log_level_api}

    def start(self):
        """Connect to and authenticate with Axonius."""
        if not self._started:
            try:
                self._auth.login()
            except Exception as exc:
                if not self._wraperror:
                    raise

                pre = f"Unable to connect to {self._http.url!r}"

                if isinstance(exc, requests.ConnectTimeout):
                    timeout = self._http.CONNECT_TIMEOUT
                    msg = f"{pre}: connection timed out after {timeout} seconds"
                    cnxexc = ConnectError(msg)
                elif isinstance(exc, requests.ConnectionError):
                    reason = self._get_exc_reason(exc=exc)
                    cnxexc = ConnectError(f"{pre}: {reason}")
                elif isinstance(exc, InvalidCredentials):
                    cnxexc = ConnectError(f"{pre}: Invalid Credentials supplied")
                else:
                    cnxexc = ConnectError(f"{pre}: {exc}")
                cnxexc.exc = exc
                raise cnxexc

            self._started = True
            LOG.info(str(self))

    @property
    def users(self):
        """Get the object for user assets API.

        Returns:
            :obj:`axonius_api_client.api.assets.Users`
        """
        self.start()
        if not hasattr(self, "_users"):
            self._users = Users(**self._api_args)
        return self._users

    @property
    def devices(self):
        """Get the object for user assets API.

        Returns:
            :obj:`axonius_api_client.assets.Devices`
        """
        self.start()
        if not hasattr(self, "_devices"):
            self._devices = Devices(**self._api_args)
        return self._devices

    @property
    def adapters(self):
        """Get the object for adapters API.

        Returns:
            :obj:`axonius_api_client.adapters.Adapters`
        """
        self.start()
        if not hasattr(self, "_adapters"):
            self._adapters = Adapters(**self._api_args)
        return self._adapters

    @property
    def enforcements(self):
        """Get the object for enforcements API.

        Returns:
            :obj:`axonius_api_client.enforcements.Enforcements`
        """
        self.start()
        if not hasattr(self, "_enforcements"):
            self._enforcements = Enforcements(**self._api_args)
        return self._enforcements

    @property
    def system(self):
        """Get the object for system API.

        Returns:
            :obj:`axonius_api_client.system.System`
        """
        self.start()
        if not hasattr(self, "_system"):
            self._system = System(**self._api_args)
        return self._system

    def __str__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        client = getattr(self, "_http", "")
        url = getattr(client, "url", self._http_args["url"])

        if self._started:
            about = self.system.meta.about()
            version = (about["Version"] or "DEMO").replace("_", ".")
            built = about["Build Date"]
            return f"Connected to {url!r} version {version} ({built})"
        else:
            return f"Not connected to {url!r}"

    def __repr__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        return self.__str__()

    @classmethod
    def _get_exc_reason(cls, exc):
        """Trim exceptions down to a more user friendly display.

        Uses :attr:`REASON_RES` to do regex substituions.

        Args:
            exc (:obj:`Exception`): Exception to trim down.

        Returns:
            :obj:`str`: prettied up str if match found, else original exception str
        """
        reason = str(exc)
        for reason_re in cls.REASON_RES:
            if reason_re.search(reason):
                return reason_re.sub(r"\1", reason).rstrip("')")
        return reason



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ...api.parsers.tables import tablize_adapters
from ...tools import json_dump
from ..context import CONTEXT_SETTINGS, click
from ..options import AUTH, add_options

OPTIONS = [
    *AUTH,
    click.option(
        "--export-format",
        "-xf",
        "export_format",
        type=click.Choice(["json-full", "json", "table", "str", "str-args"]),
        help="Format of to export data in",
        default="str",
        show_envvar=True,
        show_default=True,
    ),
]


@click.command(name="get", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, export_format, **kwargs):
    """Get adapter information."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        rows = client.adapters.get(**kwargs)

    if export_format == "json":

        basic_keys = [
            "name",
            "node_name",
            "cnx_count_total",
            "cnx_count_broken",
            "cnx_count_working",
        ]

        basic = []
        for row in rows:
            new_row = {k: row[k] for k in basic_keys}
            new_row["cnx_ids_broken"] = [x["id"] for x in row["cnx"] if not x["working"]]
            new_row["cnx_ids_working"] = [x["id"] for x in row["cnx"] if x["working"]]
            basic.append(new_row)

        click.secho(json_dump(basic))
    elif export_format == "json-full":
        click.secho(json_dump(rows))
    elif export_format == "table":
        click.secho(tablize_adapters(adapters=rows))
    elif export_format == "str-args":
        lines = "\n".join(
            ["--node-name {node_name} --name {name}".format(**row) for row in rows]
        )
        click.secho(lines)
    elif export_format == "str":
        lines = "\n".join(["{node_name}:{name}".format(**row) for row in rows])
        click.secho(lines)
    ctx.exit(0)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ...context import CONTEXT_SETTINGS, click
from ...options import AUTH, add_options
from .grp_common import EXPORT, handle_export

OPTIONS = [
    *AUTH,
    EXPORT,
]


@click.command(name="get-default", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, **kwargs):
    """Get the default role for external users."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        data = client.system.roles.get_default()

    handle_export(ctx=ctx, data=data, **kwargs)



# -*- coding: utf-8 -*-
"""API model for working with system configuration."""
import math

from ...tools import dt_now, dt_parse, timedelta
from ..mixins import ChildMixins


class Discover(ChildMixins):
    """Child API model for working with discovery cycles."""

    def get(self):
        """Get lifecycle metadata.

        Returns:
            :obj:`dict`: discovery cycle metadata
        """
        return parse_lifecycle(raw=self._get())

    @property
    def is_running(self):
        """Check if discovery cycle is running.

        Returns:
            :obj:`bool`: if discovery cycle is running
        """
        return self.get()["is_running"]

    def start(self):
        """Start a discovery cycle if one is not running.

        Returns:
            :obj:`dict`: discovery cycle metadata
        """
        if not self.is_running:
            self._start()
        return self.get()

    def stop(self):
        """Stop a discovery cycle if one is running.

        Returns:
            :obj:`dict`: discovery cycle metadata
        """
        if self.is_running:
            self._stop()
        return self.get()

    def _get(self):
        """Direct API method to get discovery cycle metadata.

        Returns:
            :obj:`dict`: discovery cycle metadata
        """
        path = self.router.discover_lifecycle
        return self.request(method="get", path=path)

    def _start(self):
        """Direct API method to start a discovery cycle."""
        path = self.router.discover_start
        return self.request(method="post", path=path)

    def _stop(self):
        """Direct API method to stop a discovery cycle.

        Returns:
            :obj:`dict`: discovery cycle metadata
        """
        path = self.router.discover_stop
        return self.request(method="post", path=path)


def parse_lifecycle(raw):
    """Pass."""
    parsed = {}

    finish_dt = raw["last_finished_time"]
    start_dt = raw["last_start_time"]

    if finish_dt:
        finish_dt = dt_parse(finish_dt)
    if start_dt:
        start_dt = dt_parse(start_dt)

    if (finish_dt and start_dt) and finish_dt >= start_dt:
        took_seconds = (finish_dt - start_dt).seconds
        took_minutes = math.ceil(took_seconds / 60)
    else:
        took_minutes = -1

    next_seconds = raw["next_run_time"]
    next_minutes = math.ceil(next_seconds / 60)
    next_dt = dt_now() + timedelta(seconds=next_seconds)

    parsed["last_start_date"] = str(start_dt)
    parsed["last_finish_date"] = str(finish_dt)
    parsed["last_took_minutes"] = took_minutes

    parsed["next_start_date"] = str(next_dt)
    parsed["next_in_minutes"] = next_minutes

    parsed["is_running"] = not raw["status"] == "done"
    parsed["phases_done"] = [x["name"] for x in raw["sub_phases"] if x["status"] == 1]
    parsed["phases_pending"] = [x["name"] for x in raw["sub_phases"] if x["status"] != 1]
    parsed["phases"] = [parse_sub_phase(raw=x) for x in raw["sub_phases"]]
    return parsed


def parse_sub_phase(raw):
    """Pass."""
    parsed = {}
    parsed["is_done"] = raw["status"] == 1
    parsed["name"] = raw["name"]
    parsed["progress"] = {}
    for name, status in raw["additional_data"].items():
        if status not in parsed["progress"]:
            parsed["progress"][status] = []
        parsed["progress"][status].append(name)
    return parsed



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import click

from ...context import AliasedGroup
from . import (cmd_get, cmd_get_section, cmd_get_subsection,
               cmd_update_section, cmd_update_section_from_json,
               cmd_update_subsection, cmd_update_subsection_from_json)


@click.group(cls=AliasedGroup)
def settings_core():
    """Group: Global Settings."""


@click.group(cls=AliasedGroup)
def settings_lifecycle():
    """Group: Lifecycle Settings."""


@click.group(cls=AliasedGroup)
def settings_gui():
    """Group: GUI Settings."""


CMDS = [
    cmd_get,
    cmd_get_section,
    cmd_get_subsection,
    cmd_update_section,
    cmd_update_subsection,
    cmd_update_section_from_json,
    cmd_update_subsection_from_json,
]

for cmd in CMDS:
    settings_core.add_command(cmd.cmd)
    settings_lifecycle.add_command(cmd.cmd)
    settings_gui.add_command(cmd.cmd)



# -*- coding: utf-8 -*-
"""Conf for py.test."""
import os

import dotenv
import pytest
from axonius_api_client.api import enforcements, system
from axonius_api_client.api.adapters import Adapters
from axonius_api_client.api.adapters.cnx import Cnx
from axonius_api_client.api.assets import (Devices, Users, fields, labels,
                                           saved_query)
from axonius_api_client.constants import CSV_ADAPTER, DEFAULT_NODE

from .meta import CSV_FILECONTENT_STR, CSV_FILENAME, QUERIES
from .utils import (check_apiobj, check_apiobj_children, check_apiobj_xref,
                    get_auth)

dotenv.load_dotenv()

AX_URL = os.environ.get("AX_URL", None) or None
AX_KEY = os.environ.get("AX_KEY", None) or None
AX_SECRET = os.environ.get("AX_SECRET", None) or None


def pytest_addoption(parser):
    """Add API connection options."""
    parser.addoption(
        "--ax-url",
        action="store",
        default=AX_URL,
        required=not bool(AX_URL),
        help="URL of Axonius API",
    )
    parser.addoption(
        "--ax-key",
        action="store",
        default=AX_KEY,
        required=not bool(AX_KEY),
        help="API key for Axonius API",
    )
    parser.addoption(
        "--ax-secret",
        action="store",
        default=AX_SECRET,
        required=not bool(AX_SECRET),
        help="API secret for Axonius API",
    )


def pytest_configure(config):
    """Ini file additions."""
    config.addinivalue_line(
        "filterwarnings", "error::axonius_api_client.exceptions.AxonWarning"
    )
    config.addinivalue_line(
        "filterwarnings", "ignore::urllib3.exceptions.InsecureRequestWarning"
    )


@pytest.fixture(scope="session")
def api_devices(request):
    """Get a fully loaded Devices API object."""
    auth = get_auth(request)

    obj = Devices(auth=auth)
    assert isinstance(obj.fields_default, list)

    check_apiobj(authobj=auth, apiobj=obj)

    check_apiobj_children(
        apiobj=obj,
        labels=labels.Labels,
        saved_query=saved_query.SavedQuery,
        fields=fields.Fields,
    )

    check_apiobj_xref(apiobj=obj, adapters=Adapters)

    field_complex = "specific_data.data.network_interfaces"
    cb_assets_query = QUERIES["exist_complex"].format(f=field_complex)

    obj.TEST_DATA = getattr(obj, "TEST_DATA", {})
    obj.TEST_DATA["field_complex"] = field_complex
    obj.TEST_DATA["field_simple"] = "specific_data.data.public_ips"
    obj.TEST_DATA["cb_assets_query"] = cb_assets_query
    obj.TEST_DATA["field_complexes"] = obj.fields_default + [field_complex]
    obj.TEST_DATA["field_main"] = obj.FIELD_HOSTNAME
    return obj


@pytest.fixture(scope="session")
def api_users(request):
    """Get a fully loaded Users API object."""
    auth = get_auth(request)

    obj = Users(auth=auth)

    assert isinstance(obj.fields_default, list)

    check_apiobj(authobj=auth, apiobj=obj)

    check_apiobj_children(
        apiobj=obj,
        labels=labels.Labels,
        saved_query=saved_query.SavedQuery,
        fields=fields.Fields,
    )

    check_apiobj_xref(apiobj=obj, adapters=Adapters)

    field_complex = "specific_data.data.associated_devices"
    cb_assets_query = QUERIES["exist_complex"].format(f=field_complex)

    obj.TEST_DATA = getattr(obj, "TEST_DATA", {})
    obj.TEST_DATA["field_complex"] = field_complex
    obj.TEST_DATA["field_simple"] = "specific_data.data.user_sid"
    obj.TEST_DATA["cb_assets_query"] = cb_assets_query
    obj.TEST_DATA["field_complexes"] = obj.fields_default
    obj.TEST_DATA["field_main"] = obj.FIELD_USERNAME
    return obj


@pytest.fixture(scope="session")
def api_enforcements(request):
    """Pass."""
    auth = get_auth(request)
    obj = enforcements.Enforcements(auth=auth)
    check_apiobj(authobj=auth, apiobj=obj)
    return obj


@pytest.fixture(scope="session")
def api_run_action(request):
    """Pass."""
    auth = get_auth(request)
    obj = enforcements.actions.RunAction(auth=auth)
    check_apiobj(authobj=auth, apiobj=obj)
    return obj


@pytest.fixture(scope="session")
def api_adapters(request):
    """Pass."""
    auth = get_auth(request)
    obj = Adapters(auth=auth)
    check_apiobj(authobj=auth, apiobj=obj)
    check_apiobj_children(apiobj=obj, cnx=Cnx)
    return obj


@pytest.fixture(scope="session")
def api_system(request):
    """Pass."""
    auth = get_auth(request)
    obj = system.System(auth=auth)
    check_apiobj(authobj=auth, apiobj=obj)
    check_apiobj_children(
        apiobj=obj,
        nodes=system.nodes.Nodes,
        settings_core=system.settings.SettingsCore,
        settings_gui=system.settings.SettingsGui,
        settings_lifecycle=system.settings.SettingsLifecycle,
        meta=system.meta.Meta,
        discover=system.discover.Discover,
        users=system.users.Users,
        roles=system.roles.Roles,
    )
    return obj


@pytest.fixture(scope="session")
def csv_file_path(api_adapters):
    """Pass."""
    data = api_adapters.file_upload(
        name=CSV_ADAPTER,
        node=DEFAULT_NODE,
        field_name="file_path",
        file_name=CSV_FILENAME,
        file_content=CSV_FILECONTENT_STR,
    )
    assert isinstance(data, dict)
    assert data["uuid"]
    assert data["filename"]
    return data


@pytest.fixture(scope="session")
def csv_file_path_broken(api_adapters):
    """Pass."""
    data = api_adapters.file_upload(
        name=CSV_ADAPTER,
        node=DEFAULT_NODE,
        field_name="file_path",
        file_name="BADWOLF",
        file_content="BADWOLF",
    )
    assert isinstance(data, dict)
    assert data["uuid"]
    assert data["filename"]
    return data



# -*- coding: utf-8 -*-
"""API model for working with system configuration."""
from ..mixins import ModelMixins
from ..routers import API_VERSION
from .central_core import CentralCore
from .discover import Discover
from .meta import Meta
from .nodes import Nodes
from .roles import Roles
from .settings import SettingsCore, SettingsGui, SettingsLifecycle
from .users import Users


class System(ModelMixins):
    """System methods."""

    @property
    def router(self):
        """Router for this API model.

        Returns:
            :obj:`.routers.Router`: REST API route defs
        """
        return API_VERSION.system

    def _init(self, auth, **kwargs):
        """Post init method for subclasses to use for extra setup.

        Args:
            auth (:obj:`.auth.Model`): object to use for auth and sending API requests
        """
        self.nodes = Nodes(parent=self)
        self.settings_core = SettingsCore(parent=self)
        self.settings_gui = SettingsGui(parent=self)
        self.settings_lifecycle = SettingsLifecycle(parent=self)
        self.meta = Meta(parent=self)
        self.discover = Discover(parent=self)
        self.users = Users(parent=self)
        self.roles = Roles(parent=self)
        self.central_core = CentralCore(parent=self)
        super(System, self)._init(auth=auth, **kwargs)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ...context import CONTEXT_SETTINGS, click
from ...options import AUTH, add_options
from .grp_common import EXPORT, handle_export

# ROLE_NAME = click.option(
#     "--name",
#     "-n",
#     "name",
#     help="Name of role to get",
#     default=None,
#     required=False,
#     show_envvar=True,
#     show_default=True,
# )

# XXX MISSING: adapters cnx update-by-id

OPTIONS = [
    *AUTH,
    EXPORT,
    # ROLE_NAME,
]

f"{OPTIONS}"


@click.command(name="get", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, **kwargs):
    """Get all roles."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        data = client.system.roles.get()

    handle_export(ctx=ctx, data=data, **kwargs)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ....tools import json_dump
from ...context import CONTEXT_SETTINGS, click
from ...options import AUTH, add_options

KEY_NAME = click.option(
    "--key-name",
    "-kn",
    "key_name",
    help="Key name of file object in [bucket_name] to restore",
    required=True,
    show_envvar=True,
    show_default=True,
)

BUCKET_NAME = click.option(
    "--bucket-name",
    "-bn",
    "bucket_name",
    default=None,
    help="Name of bucket in S3 to get [key_name] from",
    show_envvar=True,
    show_default=True,
)

ACCESS_KEY_ID = click.option(
    "--access-key-id",
    "-aki",
    "access_key_id",
    default=None,
    help="AWS Access Key Id to use to access [bucket_name]",
    show_envvar=True,
    show_default=True,
)
SECRET_ACCESS_KEY = click.option(
    "--secret-access-key",
    "-sak",
    "secret_access_key",
    default=None,
    help="AWS Secret Access Key to use to access [bucket_name]",
    show_envvar=True,
    show_default=True,
)
PRESHARED_KEY = click.option(
    "--preshared-key",
    "-pk",
    "preshared_key",
    default=None,
    help="Password to use to decrypt [key_name]",
    show_envvar=True,
    show_default=True,
)

ALLOW_RE_RESTORE = click.option(
    "--allow-re-restore/--no-allow-re-restore",
    "-arr/-narr",
    "allow_re_restore",
    help="Restore [key_name] even if it has already been restored",
    is_flag=True,
    default=False,
    show_envvar=True,
    show_default=True,
)
DELETE_BACKUPS = click.option(
    "--delete-backups/--no-delete-backups",
    "-db/-ndb",
    "delete_backups",
    help="Delete [key_name] from [bucket_name] after restore has finished",
    is_flag=True,
    default=None,
    show_envvar=True,
    show_default=True,
)

OPTIONS = [
    *AUTH,
    ACCESS_KEY_ID,
    SECRET_ACCESS_KEY,
    PRESHARED_KEY,
    ALLOW_RE_RESTORE,
    DELETE_BACKUPS,
    BUCKET_NAME,
    KEY_NAME,
]

EPILOG = """
If values for these options are not provided, they will default to
the settings under Global Settings > Amazon S3 Settings:

\b
  * bucket-name: Amazon S3 bucket name
  * access-key-id: AWS Access Key Id
  * secret-access-key: AWS Secret Access Key
  * preshared-key: Backup encryption passphrase

"""


@click.command(
    name="restore-from-aws-s3", context_settings=CONTEXT_SETTINGS, epilog=EPILOG,
)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, **kwargs):
    """Perform a manual restore of a backup in AWS S3."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        data = client.system.central_core.restore_from_aws_s3(**kwargs)

    click.secho(json_dump(data))



# -*- coding: utf-8 -*-
"""Test suite."""

import pytest
from axonius_api_client.constants import DEFAULT_PERM
from axonius_api_client.exceptions import ApiError, NotFoundError

from ...meta import TEST_PERM, TEST_ROLE


@pytest.mark.skip(f"Waiting for update to 3.3!")  # XXX update public API for roles/users
class TestSystemRoles:
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_system):
        """Pass."""
        return api_system.roles

    def test__get(self, apiobj):
        """Pass."""
        data = apiobj._get()
        assert isinstance(data, list)
        for x in data:
            assert isinstance(x, dict)

    def test_get_set_default(self, apiobj):
        """Pass."""
        roles = apiobj.get()
        current_role = apiobj.get_default()
        assert isinstance(current_role, dict)

        new_roles = [x for x in roles if x["name"] != current_role["name"]]
        if new_roles:
            new_role = new_roles[0]["name"]
            updated_role = apiobj.set_default(new_role)
            assert isinstance(updated_role, dict)
            assert updated_role["name"] == new_role

    def test_add_invalid_perm(self, apiobj):
        """Pass."""
        with pytest.raises(ApiError):
            apiobj.add("x", "badwolf_itai")

    def test_add_get_update_delete(self, apiobj):
        """Pass."""
        try:
            apiobj.get(name=TEST_ROLE)
        except NotFoundError:
            pass
        else:
            apiobj.delete(name=TEST_ROLE)

        added = apiobj.add(name=TEST_ROLE)
        assert isinstance(added, dict)
        assert added["name"] == TEST_ROLE
        assert isinstance(added["permissions"], dict)

        with pytest.raises(ApiError):
            apiobj.add(name=TEST_ROLE)

        added_permissions = added["permissions"]
        update_permissions = {}

        for k, v in added_permissions.items():
            assert v == DEFAULT_PERM
            update_permissions[k.lower()] = TEST_PERM

        updated = apiobj.update(name=TEST_ROLE, **update_permissions)
        assert isinstance(updated, dict)

        updated_permissions = updated["permissions"]
        for k, v in updated_permissions.items():
            assert v == TEST_PERM

        deleted = apiobj.delete(name=TEST_ROLE)
        assert isinstance(deleted, list)
        assert not [x for x in deleted if x["name"] == added["name"]]

        with pytest.raises(NotFoundError):
            apiobj.update(name=TEST_ROLE)

        with pytest.raises(NotFoundError):
            apiobj.delete(name=TEST_ROLE)

        with pytest.raises(NotFoundError):
            apiobj.get(name=TEST_ROLE)



# -*- coding: utf-8 -*-
"""API model for unauthenticated endpoints."""
from ..constants import LOG_LEVEL_API
from ..exceptions import ResponseNotOk
from ..http import Http
from ..logs import get_obj_log


class Entry:
    """API model for unauthenticated endpoints."""

    @property
    def is_signed_up(self):
        """Check if signup process has been done."""
        return self._signup_get()["signup"]

    def signup(self, password, company_name, contact_email):
        """Perform the initial signup."""
        # XXX 400, signup already completed
        response = self._signup_post(
            password=password, company_name=company_name, contact_email=contact_email
        )

        status = response.get("status")
        message = response.get("message")
        if status == "error":
            raise ResponseNotOk(f"{message}")
        return response

    @property
    def api_version(self):
        """Get the API version."""
        return self._api_version

    def _signup_get(self):
        """Get the status of initial signup."""
        response = self.http(method="get", path="api", route="signup_api")
        return response.json()

    def _signup_post(self, password, company_name, contact_email):
        """Do the initial signup."""
        data = {
            "companyName": company_name,
            "contactEmail": contact_email,
            "userName": "admin",
            "newPassword": password,
            "confirmNewPassword": password,
        }
        response = self.http(method="post", path="api", route="signup_api", json=data)
        return response.json()

    def _api_version(self):
        """Get the API version."""
        response = self.http(method="get", path="api", route="api")
        return response.text

    def __init__(self, url, **kwargs):
        """Pass."""
        log_level = kwargs.get("log_level", LOG_LEVEL_API)
        self.LOG = get_obj_log(obj=self, level=log_level)
        kwargs.setdefault("certwarn", False)
        self.http = Http(url=url, **kwargs)



# -*- coding: utf-8 -*-
"""Constants."""
import logging
import os
import pathlib
import sys

import dotenv

from . import __package__ as PACKAGE_ROOT

DEFAULT_PATH = os.getcwd()

AX_ENV = os.environ.get("AX_ENV", "").strip() or DEFAULT_PATH


def load_dotenv(ax_env=AX_ENV, reenv=False, verbose=False):
    """Pass."""
    if reenv:
        ax_env = os.environ.get("AX_ENV", "").strip() or DEFAULT_PATH

    ax_env_path = pathlib.Path(ax_env).expanduser().resolve()

    if ax_env_path.is_dir():
        ax_env_path = ax_env_path / ".env"

    return (
        dotenv.load_dotenv(dotenv_path=str(ax_env_path), verbose=verbose),
        ax_env_path,
    )


load_dotenv()

PY36 = sys.version_info[0:2] >= (3, 6)
""":obj:`bool`: python version is 3.6 or higher"""

PY37 = sys.version_info[0:2] >= (3, 7)
""":obj:`bool`: python version is 3.7 or higher"""

COMPLEX = (dict, list, tuple)
""":obj:`tuple` of :obj:`type`: types that are considered as complex."""

SIMPLE = (str, int, bool, float)
""":obj:`tuple` of :obj:`type`: types that are considered as simple"""

EMPTY = [None, "", [], {}, ()]
""":obj:`list` of :obj:`type`: values that should be considered as empty"""

YES = [True, 1, "1", "true", "t", "yes", "y"]
""":obj:`list` of :obj:`type`: values that should be considered as truthy"""

NO = [False, 0, "0", "false", "f", "no", "n"]
""":obj:`list` of :obj:`type`: values that should be considered as falsey"""

MAX_PAGE_SIZE = 2000
""":obj:`int`: maximum page size that REST API allows"""

PAGE_SIZE = MAX_PAGE_SIZE
PAGE_SLEEP = 0
PAGE_CACHE = False

GUI_PAGE_SIZES = [25, 50, 100]
""":obj:`list` of :obj:`int`: valid page sizes for GUI paging"""

LOG_REQUEST_ATTRS_BRIEF = [
    "request to {request.url!r}",
    "method={request.method!r}",
    "size={size}",
]
""":obj:`list` of :obj:`str`: request attributes to log when verbose=False"""

RESPONSE_ATTR_MAP = {
    "url": "{response.url!r}",
    "size": "{response.body_size}",
    "method": "{response.request.method!r}",
    "status": "{response.status_code!r}",
    "reason": "{response.reason!r}",
    "elapsed": "{response.elapsed}",
    "headers": "{response.headers}",
}

REQUEST_ATTR_MAP = {
    "url": "{request.url!r}",
    "size": "{request.body_size}",
    "method": "{request.method!r}",
    "headers": "{request.headers}",
}

TIMEOUT_CONNECT = 5
""":obj:`int`: seconds to wait for connection to API."""

TIMEOUT_RESPONSE = 900
""":obj:`int`: seconds to wait for response from API."""

LOG_FMT_VERBOSE = "%(asctime)s %(levelname)-8s [%(name)s:%(funcName)s()] %(message)s"
LOG_FMT_BRIEF = "%(levelname)-8s [%(name)s] %(message)s"

DEBUG = os.environ.get("AX_DEBUG", "").lower().strip()
DEBUG = any([DEBUG == x for x in YES])

LOG_FMT_CONSOLE = LOG_FMT_VERBOSE if DEBUG else LOG_FMT_BRIEF
""":obj:`str`: default logging format to use for console logs"""

LOG_FMT_FILE = LOG_FMT_VERBOSE
""":obj:`str`: default logging format to use for file logs"""

LOG_DATEFMT_CONSOLE = "%m/%d/%Y %I:%M:%S %p"
""":obj:`str`: default datetime format to use for console logs"""

LOG_DATEFMT_FILE = "%m/%d/%Y %I:%M:%S %p"
""":obj:`str`: default datetime format to use for file logs"""

LOG_LEVEL_CONSOLE = "debug"
""":obj:`str`: default logging level to use for console log handlers"""

LOG_LEVEL_FILE = "debug"
""":obj:`str`: default logging level to use for file log handlers"""

LOG_LEVEL_HTTP = "debug"
""":obj:`str`: default logging level to use for :obj:`axonius_api_client.http.Http`"""

LOG_LEVEL_AUTH = "debug"
""":obj:`str`: default logging level to use for :obj:`axonius_api_client.auth.Mixins`"""

LOG_LEVEL_API = "debug"
""":obj:`str`: default logging level to use for
:obj:`axonius_api_client.api.mixins.Mixins`"""

LOG_LEVEL_PACKAGE = "debug"
""":obj:`str`: default logging level to use for :mod:`axonius_api_client`"""

LOG_LEVELS_STR = ["debug", "info", "warning", "error", "fatal"]
""":obj:`list` of :obj:`str`: valid logging level strs"""

LOG_LEVELS_STR_CSV = ", ".join(LOG_LEVELS_STR)
""":obj:`str`: csv of valid logging level strs"""

LOG_LEVELS_INT = [getattr(logging, x.upper()) for x in LOG_LEVELS_STR]
""":obj:`list` of :obj:`int`: valid logging level ints"""

LOG_LEVELS_INT_CSV = ", ".join([str(x) for x in LOG_LEVELS_INT])
""":obj:`str`: csv of valid logging level ints"""

LOG_FILE_PATH = DEFAULT_PATH
""":obj:`str`: default path to use for log files"""

LOG_FILE_PATH_MODE = 0o700
""":obj:`str`: default permisisons to use when creating directories"""

LOG_FILE_NAME = f"{PACKAGE_ROOT}.log"
""":obj:`str`: default log file name to use"""

LOG_FILE_MAX_MB = 5
""":obj:`int`: default rollover trigger in MB"""

LOG_FILE_MAX_FILES = 5
""":obj:`int`: default max rollovers to keep"""

LOG_NAME_STDERR = "handler_stderr"
""":obj:`str`: default handler name to use for STDERR log"""

LOG_NAME_STDOUT = "handler_stdout"
""":obj:`str`: default handler name to use for STDOUT log"""

LOG_NAME_FILE = "handler_file"
""":obj:`str`: default handler name to use for file log"""

SETTING_UNCHANGED = ["unchanged"]
""":obj:`list` of :obj:`str`: ref used by REST API when supplying a password
field that should remain the same as what is already in the database"""

DEFAULT_NODE = "Master"
""":obj:`str`: default node name to use"""

CSV_ADAPTER = "csv"
""":obj:`str`: name of csv adapter"""
CSV_FIELD_NAME = "file_path"
CNX_SANE_DEFAULTS = {
    "all": {"verify_ssl": False},
    "csv": {
        "is_users": False,
        "is_installed_sw": False,
        "s3_use_ec2_attached_instance_profile": False,
        "verify_ssl": False,
    },
}
DEFAULT_PERM = "ReadOnly"
""":obj:`str`: default user permission to use"""

VALID_PERMS = ["Restricted", "ReadWrite", "ReadOnly"]
""":obj:`list` of :obj:`str`: valid user permissions"""
PERM_SETS = [
    "Adapters",
    "Dashboard",
    "Devices",
    "Enforcements",
    "Instances",
    "Reports",
    "Settings",
    "Users",
]

FIELD_TRIM_LEN = 32000
FIELD_TRIM_STR = "...TRIMMED - {field_len} characters over {trim_len}"
FIELD_JOINER = "\n"
TABLE_FORMAT = "fancy_grid"
TABLE_MAX_ROWS = 5

OK_ARGS = {"fg": "green", "bold": True, "err": True}

OK_TMPL = "** {msg}"

WARN_ARGS = {"fg": "yellow", "bold": True, "err": True}

WARN_TMPL = "** WARNING: {msg}"

ERROR_ARGS = {"fg": "red", "bold": True, "err": True}

ERROR_TMPL = "** ERROR: {msg}"
AGG_ADAPTER_NAME = "agg"
AGG_ADAPTER_TITLE = "Aggregated"
ALL_NAME = "all"
AGG_ADAPTER_ALTS = ["generic", "general", "specific", "agg", "aggregated"]
""":obj:`list` of :obj:`str`: list of alternatives for 'generic' adapter."""


NORM_TYPE_MAP = (
    # (type, format, items.type, items.format), normalized
    (("string", "", "", ""), "string"),
    (("string", "date-time", "", ""), "string_datetime"),
    (("string", "image", "", ""), "string_image"),
    (("string", "version", "", ""), "string_version"),
    (("string", "ip", "", ""), "string_ipaddress"),
    (("bool", "", "", ""), "bool"),
    (("integer", "", "", ""), "integer"),
    (("number", "", "", ""), "number"),
    (("array", "table", "array", ""), "complex_table"),
    (("array", "", "array", ""), "complex"),
    (("array", "", "integer", ""), "list_integer"),
    (("array", "", "string", ""), "list_string"),
    (("array", "", "string", "tag"), "list_string"),
    (("array", "version", "string", "version"), "list_string_version"),
    (("array", "date-time", "string", "date-time"), "list_string_datetime"),
    (("array", "subnet", "string", "subnet"), "list_string_subnet"),
    (("array", "discrete", "string", "logo"), "list_string"),
    (("array", "ip", "string", "ip"), "list_string_ipaddress"),
)


GET_SCHEMAS_KEYS = ["name", "name_qual", "name_base"]
GET_SCHEMA_KEYS = ["name_base", "name_qual", "name"]

SCHEMAS_CUSTOM = {
    "report_adapters_missing": {
        "adapters_missing": {
            "adapter_name": "report",
            "column_name": "report:adapters_missing",
            "column_title": "Report: Adapters Missing",
            "is_complex": False,
            "is_list": True,
            "is_root": True,
            "parent": "root",
            "name": "adapters_missing",
            "name_base": "adapters_missing",
            "name_qual": "adapters_missing",
            "title": "Adapters Missing",
            "type": "string",
            "type_norm": "list_string",
            "is_custom": True,
        }
    }
}

MAX_BODY_LEN = 2000
GENERIC_NAME = "AdapterBase"


KEY_MAP_CNX = [
    ("adapter_name", "Adapter", 0),
    ("node_name", "Node", 0),
    ("id", "ID", 0),
    ("uuid", "UUID", 0),
    ("working", "Working", 0),
    ("error", "Error", 20),
    ("label", "Label", 0),
    ("schemas", None, 0),
]

KEY_MAP_ADAPTER = [
    ("name", "Name", 0),
    ("node_name", "Node", 0),
    ("cnx_count_total", "Connections", 0),
    ("cnx_count_broken", "Broken", 0),
    ("cnx_count_working", "Working", 0),
]

KEY_MAP_SCHEMA = [
    ("name", "Name", 0),
    ("title", "Title", 30),
    ("type", "Type", 0),
    ("required", "Required", 0),
    ("default", "Default", 0),
    ("description", "Description", 20),
    ("format", "Format", 0),
]
CNX_GONE = "Server is already gone, please try again after refreshing the page"
CNX_RETRY = 15



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ....tools import json_dump
from ...context import CONTEXT_SETTINGS, click
from ...options import (AUTH, FIELDS_SELECT, QUERY, add_options,
                        get_option_fields_default)
from .grp_common import SQ_OPTS

EPILOG = """Note:

Creating a saved query using this method will not create a proper set of
expressions, which will mean the query wizard will not show the filter lines
for the supplied query!
"""

OPTIONS = [
    *AUTH,
    *QUERY,
    *FIELDS_SELECT,
    get_option_fields_default(default=False),
    *SQ_OPTS,
    click.option(
        "--name",
        "-n",
        "name",
        help="Name of saved query",
        required=True,
        show_envvar=True,
        show_default=True,
    ),
]


@click.command(name="add", context_settings=CONTEXT_SETTINGS, epilog=EPILOG)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, **kwargs):
    """Add a saved query."""
    column_filters = kwargs.get("column_filters", [])
    if column_filters:
        kwargs["column_filters"] = dict(kwargs.get("column_filters", []))

    query_file = kwargs.pop("query_file", None)
    if query_file:
        kwargs["query"] = query_file.read().strip()

    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    p_grp = ctx.parent.parent.command.name
    apiobj = getattr(client, p_grp)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        row = apiobj.saved_query.add(**kwargs)

    ctx.obj.echo_ok(f"Successfully created saved query: {row['name']}")

    click.secho(json_dump(row))

    ctx.exit(0)



# -*- coding: utf-8 -*-
"""Test suite for axonius_api_client.tools."""
from ....cli import cli
from ....constants import CSV_ADAPTER, DEFAULT_NODE
from ....tools import json_dump, json_load
from ...utils import load_clirunner


class TestGrpAdaptersCmdConfigGetUpdate:
    """Pass."""

    def test_json_full(self, request, monkeypatch):
        """Pass."""
        runner = load_clirunner(request, monkeypatch)

        args1 = [
            "adapters",
            "config-get",
            "--name",
            CSV_ADAPTER,
            "--node-name",
            DEFAULT_NODE,
            "--export-format",
            "json-full",
        ]
        result1 = runner.invoke(cli=cli, args=args1)

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0

        json1 = json_load(stdout1)
        assert isinstance(json1, dict)

    def test_json_basic(self, request, monkeypatch):
        """Pass."""
        runner = load_clirunner(request, monkeypatch)

        args1 = [
            "adapters",
            "config-get",
            "--name",
            CSV_ADAPTER,
            "--node-name",
            DEFAULT_NODE,
            "--export-format",
            "json",
        ]
        result1 = runner.invoke(cli=cli, args=args1)

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0

        json1 = json_load(stdout1)
        assert isinstance(json1, dict)

        key = "user_last_fetched_threshold_hours"
        old_value = json1.get(key, 48)
        new_value = old_value + 1

        args2 = [
            "adapters",
            "config-update",
            "--name",
            CSV_ADAPTER,
            "--node-name",
            DEFAULT_NODE,
            "--export-format",
            "json",
            "--config",
            f"{key}={new_value}",
        ]
        result2 = runner.invoke(cli=cli, args=args2)

        stderr2 = result2.stderr
        stdout2 = result2.stdout
        exit_code2 = result2.exit_code

        assert stdout2
        assert stderr2
        assert exit_code2 == 0

        json2 = json_load(stdout2)
        assert isinstance(json2, dict)
        assert json2[key] == new_value

        json2[key] = old_value

        args3 = [
            "adapters",
            "config-update-from-json",
            "--name",
            CSV_ADAPTER,
            "--node-name",
            DEFAULT_NODE,
            "--export-format",
            "json",
            "--input-file",
            "-",
        ]
        result3 = runner.invoke(cli=cli, args=args3, input=json_dump(json2))

        stderr3 = result3.stderr
        stdout3 = result3.stdout
        exit_code3 = result3.exit_code

        assert stdout3
        assert stderr3
        assert exit_code3 == 0

        json3 = json_load(stdout3)
        assert isinstance(json3, dict)
        assert json3[key] == old_value



# -*- coding: utf-8 -*-
"""API models package."""
from . import (central_core, discover, meta, nodes, roles, settings, system,
               users)
from .system import System

__all__ = (
    "System",
    "system",
    "discover",
    "nodes",
    "meta",
    "roles",
    "settings",
    "users",
    "central_core",
)



# -*- coding: utf-8 -*-
"""API models for working with device and user assets."""
import json

import jsonstreams

from ...tools import listify
from .base import Base


class Json(Base):
    """Pass."""

    CB_NAME = "json"

    def start(self, **kwargs):
        """Create jsonstream and associated file descriptor."""
        super(Json, self).start(**kwargs)
        self.open_fd()
        self._stream = JsonStream(fd=self._fd)

    def stop(self, **kwargs):
        """Close jsonstream and associated file descriptor."""
        super(Json, self).stop(**kwargs)
        self.do_export_schema()
        self._stream.close()
        self._fd.write("\n")
        self.close_fd()

    def do_export_schema(self):
        """Pass."""
        export_schema = self.GETARGS.get("export_schema", False)
        if export_schema:
            self._stream.write({"schemas": self.schemas_final()})

    def row(self, row):
        """Write row to jsonstreams and delete it."""
        return_row = [{"internal_axon_id": row["internal_axon_id"]}]

        rows = super(Json, self).row(row=row)

        for row_new in listify(rows):
            self._stream.write(row_new)
            del row_new

        del rows
        del row

        return return_row


class JsonStream:
    """Wrap jsonstreams.Stream object."""

    def __init__(self, fd, **kwargs):
        """Wrap jsonstreams.Stream object."""
        self.__fd = fd

        encoder = kwargs.get("encoder", json.JSONEncoder)
        indent = kwargs.get("indent", 2)

        self.__inst = jsonstreams.Array(
            fd=self.__fd,
            indent=indent,
            baseindent=0,
            encoder=encoder(indent=indent),
            pretty=kwargs.get("pretty", True),
        )

        self.subobject = self.__inst.subobject
        self.subarray = self.__inst.subarray

    def close(self):
        """Close the root element and print a message."""
        self.__inst.close()

    def write(self, *args, **kwargs):
        """Write values into the stream."""
        self.__inst.write(*args, **kwargs)



# -*- coding: utf-8 -*-
"""API model for working with system configuration."""
from ..mixins import ChildMixins


class CentralCore(ChildMixins):
    """Child API model for working with instance metadata."""

    def get(self):
        """Get the current central core configuration."""
        return self._get()

    def update(self, enabled, delete_backups):
        """Update the current central core configuration."""
        return self._update(enabled=enabled, delete_backups=delete_backups)

    def restore_from_aws_s3(
        self,
        key_name,
        bucket_name=None,
        access_key_id=None,
        secret_access_key=None,
        preshared_key=None,
        allow_re_restore=False,
        delete_backups=False,
    ):
        """Perform a restore from a file object in an AWS S3 Bucket."""
        restore_opts = {}
        restore_opts["key_name"] = key_name
        restore_opts["allow_re_restore"] = allow_re_restore

        if delete_backups is not None:
            restore_opts["delete_backups"] = delete_backups

        if bucket_name is not None:
            restore_opts["bucket_name"] = bucket_name

        if access_key_id is not None:
            restore_opts["access_key_id"] = access_key_id

        if secret_access_key is not None:
            restore_opts["secret_access_key"] = secret_access_key

        if preshared_key is not None:
            restore_opts["preshared_key"] = preshared_key

        return self._restore(restore_type="aws", restore_opts=restore_opts)

    def _get(self):
        """Get the current central core configuration."""
        path = self.router.central_core
        response = self.request(method="get", path=path)
        return response

    def _update(self, enabled, delete_backups):
        """Set the current central core configuration."""
        data = {"enabled": enabled, "delete_backups": delete_backups}
        path = self.router.central_core
        response = self.request(method="post", path=path, json=data)
        return response

    def _restore(self, restore_type, restore_opts):
        """Perform a central core restore operation."""
        data = {}
        data["restore_type"] = restore_type
        data.update(restore_opts)

        path = self.router.central_core_restore
        response = self.request(
            method="post", path=path, json=data, response_timeout=3600
        )
        return response



# -*- coding: utf-8 -*-
"""Test suite."""
import pytest

from ...meta import ABOUT_KEYS, ABOUT_KEYS_EMPTY_OK


class SystemMetaBase:
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_system):
        """Pass."""
        return api_system.meta

    def val_about(self, data):
        """Pass."""
        assert isinstance(data, dict)

        keys = ABOUT_KEYS
        empty_ok = ABOUT_KEYS_EMPTY_OK

        for key in keys:
            assert key in data
            assert isinstance(data[key], str)
            if key not in empty_ok:
                assert data[key]


class TestSystemMetaPrivate(SystemMetaBase):
    """Pass."""

    def val_entity_sizes(self, data):
        """Pass."""
        avg_document_size = data("avg_document_size")
        assert isinstance(avg_document_size, int)
        capped = data("capped")
        assert isinstance(capped, int)
        entities_last_point = data("entities_last_point")
        assert isinstance(entities_last_point, int)
        size = data("size")
        assert isinstance(size, int)
        assert not data

    def val_historical_sizes(self, data):
        """Pass."""
        disk_free = data.pop("disk_free")
        assert isinstance(disk_free, int) and disk_free

        disk_used = data.pop("disk_used")
        assert isinstance(disk_used, int) and disk_used

        entity_sizes = data.pop("entity_sizes")
        assert isinstance(entity_sizes, dict)

        users = entity_sizes.pop("Users", {})
        assert isinstance(users, dict)

        devices = entity_sizes.pop("Devices", {})
        assert isinstance(devices, dict)
        assert not entity_sizes
        assert not data

    def test_private_about(self, apiobj):
        """Pass."""
        data = apiobj._about()
        self.val_about(data)

    def test_private_historical_sizes(self, apiobj):
        """Pass."""
        data = apiobj._historical_sizes()
        self.val_historical_sizes(data)


class TestSystemMetaPublic(SystemMetaBase):
    """Pass."""

    def test_about(self, apiobj):
        """Pass."""
        data = apiobj.about()
        self.val_about(data)

    def test_historical_sizes(self, apiobj):
        """Pass."""
        data = apiobj.historical_sizes()
        assert isinstance(data["disk_free_mb"], int)
        assert isinstance(data["disk_used_mb"], int)
        assert isinstance(data["historical_sizes_devices"], dict)
        assert isinstance(data["historical_sizes_users"], dict)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import warnings

import click
import requests

from ..connect import Connect
from ..tools import echo_error, echo_ok, echo_warn, json_load

CONTEXT_SETTINGS = {"auto_envvar_prefix": "AX"}
SSLWARN_CLS = requests.urllib3.exceptions.InsecureRequestWarning
SSLWARN_MSG = """Unverified HTTPS request!

To enable certificate validation:
  * Set the variable: AX_CERTPATH=/path/to/cert_or_ca_bundle
  * Supply the option: -cp/--cert-path /path/to/cert_or_ca_bundle

To silence this message:
  * Set the variable: AX_CERTWARN=n
  * Supply the option: -ncw/--no-cert-warn
"""


class SplitEquals(click.ParamType):
    """Pass."""

    name = "split_equals"

    def convert(self, value, param, ctx):
        """Pass."""
        split = value.split("=", 1)

        if len(split) != 2:
            msg = "Need an '=' in --{p} with value {v!r}"
            msg = msg.format(p=param.name, v=value)
            self.fail(msg, param, ctx)

        return [x.strip() for x in split]


class AliasedGroup(click.Group):
    """Pass."""

    def get_command(self, ctx, cmd_name):
        """Pass."""
        rv = click.Group.get_command(self, ctx, cmd_name)

        if rv is not None:
            return rv

        matches = [x for x in self.list_commands(ctx) if x.startswith(cmd_name)]

        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])

        msg = "Too many matches for {v!r}: {matches}"
        msg = msg.format(v=cmd_name, matches=", ".join(sorted(matches)))
        ctx.fail(msg)


class exc_wrapper:
    """Pass."""

    EXCLUDES = (SystemExit, click.exceptions.Exit)

    def __init__(self, wraperror=True):
        """Pass."""
        self.wraperror = wraperror

    def __enter__(self):
        """Pass."""
        return self

    def __exit__(self, exc, value, traceback):
        """Pass."""
        if value and self.wraperror and not isinstance(value, self.EXCLUDES):
            msg = "WRAPPED EXCEPTION: {c.__module__}.{c.__name__}\n{v}"
            msg = msg.format(c=value.__class__, v=value)
            echo_error(msg)


class Context:
    """Pass."""

    exc_wrap = exc_wrapper
    QUIET = False

    def __init__(self):
        """Pass."""
        self.client = None
        self._connect_args = {}

    def __str__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        return format(self.client)

    def __repr__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        return self.__str__()

    def echo_ok(self, msg, **kwargs):
        """Pass."""
        if not getattr(self, "QUIET", False):
            echo_ok(msg=msg, **kwargs)

    def echo_error(self, msg, abort=True, **kwargs):
        """Pass."""
        echo_error(msg=msg, abort=abort, **kwargs)

    def echo_warn(self, msg, **kwargs):
        """Pass."""
        echo_warn(msg=msg, **kwargs)

    @property
    def wraperror(self):
        """Pass."""
        return self._connect_args.get("wraperror", True)

    def start_client(self, url, key, secret, echo=True, **kwargs):
        """Pass."""
        if not getattr(self, "client", None):
            connect_args = {}
            connect_args.update(self._connect_args)
            connect_args.update(kwargs)
            connect_args["url"] = url
            connect_args["key"] = key
            connect_args["secret"] = secret

            with self.exc_wrap(wraperror=self.wraperror):
                self.client = Connect(**connect_args)

                with warnings.catch_warnings(record=True) as caught_warnings:
                    self.client.start()

                for caught_warning in caught_warnings:
                    wmsg = caught_warning.message
                    is_ssl = isinstance(wmsg, SSLWARN_CLS)
                    wmsg = SSLWARN_MSG if is_ssl else wmsg
                    wmsg = format(wmsg)
                    self.echo_warn(wmsg)

            # warnings suck.
            warnings.simplefilter("ignore", SSLWARN_CLS)

            if echo:
                self.echo_ok(msg=str(self.client))

        return self.client

    def read_stream(self, stream):
        """Pass."""
        stream_name = format(getattr(stream, "name", stream))

        if stream.isatty():
            self.echo_error(msg=f"No input provided on {stream_name!r}", abort=True)

        # its STDIN with input or a file
        content = stream.read().strip()
        self.echo_ok(msg=f"Read {len(content)} bytes from {stream_name!r}")

        if not content:
            self.echo_error(msg=f"Empty content supplied to {stream_name!r}", abort=True)

        return content

    def read_stream_json(self, stream, expect):
        """Pass."""
        content = self.read_stream(stream=stream)

        try:
            content = json_load(obj=content, error=True)
        except Exception as exc:
            self.echo_error(msg=f"Invalid JSON supplied: {exc}", abort=True)

        if not isinstance(content, expect):

            self.echo_error(
                msg=f"JSON supplied is {type(content)}, required type is {expect}",
                abort=True,
            )

        return content


pass_context = click.make_pass_decorator(Context, ensure=True)



"""Test meta data."""
import time

QUERIES = {
    "not_last_seen_day": '(not (specific_data.data.last_seen >= date("NOW - 1d")))',
    "exist_complex": '((({f} == ({{"$exists":true,"$ne":[]}})) and {f} != []))',
    "exist_simple": '(({f} == ({{"$exists":true,"$ne":""}})))',
}

TEST_CLIENT_CERT_NAME = "client_cert.crt"
TEST_CLIENT_CERT = """
-----BEGIN CERTIFICATE-----
MIIEtDCCApygAwIBAgIJAPuK1/Z7X2zbMA0GCSqGSIb3DQEBDQUAMFgxCzAJBgNV
BAYTAkFVMRMwEQYDVQQIDApTb21lLVN0YXRlMRkwFwYDVQQKDBBBeG9uaXVzIENJ
IFRlc3RzMRkwFwYDVQQDDBBBeG9uaXVzQ0lUZXN0c0NBMB4XDTE5MTIyNDEzMTYx
NFoXDTM5MDIyMjEzMTYxNFowgYsxCzAJBgNVBAYTAklMMREwDwYDVQQIDAhUZWwg
QXZpdjEKMAgGA1UEBwwBLjEKMAgGA1UEEQwBLjEQMA4GA1UECgwHQXhvbml1czEM
MAoGA1UECwwDUiZEMQwwCgYDVQQDDANCb2IxIzAhBgkqhkiG9w0BCQEWFGJvYkBh
eG9uaXVzdGVzdHMubGFuMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA
t/p+x2hlsidVdoDkEnAV7S0t4DTQE4Iir8VGm/Rjb9Gv+O3T/VWPGYYm2roiSS0q
0DCNOSy8qx28+vccgdPdyflCte9/mas6dTRvXDM3nDLloqI9lQy8Tf1X3rhRDs8a
EZCmfiATUETPr7vp3+LjriHpSnc/siaUFdDxEWRIbAIR3zNDv7MJD7k16HQCK+5k
iCtJJXcelYC856vEASORJibUU/Q15KqgegM/ATf6NLem5fYO2dZhQMSj8nW8Cw2Q
pAo+uUdeDMLscREEG933zH0qDCNaGyjAY7JUqlHXEBtL2GSADNp2WGLyK4xMdTiB
HqGi05DvFKRXnBvs1YOdnQIDAQABo00wSzAxBgNVHSUEKjAoBggrBgEFBQcDAQYI
KwYBBQUHAwIGCCsGAQUFBwMDBggrBgEFBQcDBDAJBgNVHRMEAjAAMAsGA1UdDwQE
AwIF4DANBgkqhkiG9w0BAQ0FAAOCAgEAlyog4NTiu4jvEOisGtm4fWupNBaEC25B
c9+hntj2454NG3i3s5hkB65A6tFEoWWNSI74CEw/VKzecjqOzAuW4d5qYxA/jtX4
50Ws0fqzPxKvJTnhzzgOAHpiyxWrJASlp3fhK73Z2aRY6xNr8RoAXsQabj5wU3gm
kNzUZel5VVcB0Entmjcp2COqIvIm81Wumz+URE732hoC5ZnRDBAlu8zrMqptHU+4
WBgGBoQUMI3AJcE2UmZlCofsXni+jdr8ZN1j/xbu79W7xCHcQzm3//teWy2KAXp/
JfFfJ6tkO63hVGV9xGlIC0DK6Z17qy1uEwNKfu2qFvl6msCNT8AeyM5DJwEX//3S
ITxeZCPzznQx38Hth/ZOuCx5C8FT+mLv8Eoac+bC45HOckCK3wACYIPUCzwSDzxn
2AWBWP9ltF5/wJu854s23ylJ0Wu2NJArq5vuuA8SSf2Hb4f3Q4J6097iAQ0zunYK
t06o6EU8iDGxCZ+0Ev1Ocx/6d6wU1W/TBrN1veDQxsd5SQpNc0Q18m5YCcq1yDyt
DvpaMDAyxvLZtDAfPOd/KgpWUThIcdhAmT+tzde8bTrT306PPzbi5pIgaCsNGBb8
w4gw3wK/xMly7xZxoLD3YujN/sCcOFY3nhkB+L71GRlX8qYV5uS9fdbTKGZD7SYs
Ck4QKLYy+P0=
-----END CERTIFICATE-----
"""

TEST_CLIENT_KEY_NAME = "client_cert.key"
TEST_CLIENT_KEY = """
-----BEGIN RSA PRIVATE KEY-----
MIIEpAIBAAKCAQEAt/p+x2hlsidVdoDkEnAV7S0t4DTQE4Iir8VGm/Rjb9Gv+O3T
/VWPGYYm2roiSS0q0DCNOSy8qx28+vccgdPdyflCte9/mas6dTRvXDM3nDLloqI9
lQy8Tf1X3rhRDs8aEZCmfiATUETPr7vp3+LjriHpSnc/siaUFdDxEWRIbAIR3zND
v7MJD7k16HQCK+5kiCtJJXcelYC856vEASORJibUU/Q15KqgegM/ATf6NLem5fYO
2dZhQMSj8nW8Cw2QpAo+uUdeDMLscREEG933zH0qDCNaGyjAY7JUqlHXEBtL2GSA
DNp2WGLyK4xMdTiBHqGi05DvFKRXnBvs1YOdnQIDAQABAoIBAGOGeiDrg+AtURlL
PpYO1n24rBGW4F09UABgKwNg4I30FEsIdV6dc00uekRm3vdRHNEFAtDEN8glzT7C
gURmVZvWYNVFG3UI4RXYaMmq11GDYyBovgGsow1ZmLheY1MsjACmjLq8JVaN8wAx
GqLH/b0MkUR8YBPCtOdcYZyz8E2kohvw2P8DBmzUCFXuELHRh5eJ4nRv9ZdF8jw2
YOYnnBFrWJzE9YA+UCi67oGpMumcu91+tsenc3tvxXJX/HCSBiCc3GEHSROw0SjQ
TKLLPC+pT5wfmHGFeBPvKFpcfncx50IiTAcnHuBTQJvUvoNkiViBhS5ZG8Pranlq
w+fBlyECgYEA3wqTnkZFWR6Rl1cSUZRIAeKga380lNcnydhLbmBhiFxr7KoGiuyZ
cBcGqock0/sng6AwEl9d59ni3I5/2p4DtJmYxx8QU2XmUXSKdGH34qIXL6T27ow0
292vYIUFAT4vyawkKtO7FlKc+p5frL+dzIDcNOJPj0wfMivRDd33dokCgYEA0yo2
X9hLqNoq5X+UUVfhlgAswb7wDelx19up3QWX12eRMnwztTp070L/aoMpJaQuXUTC
HwF4Muffe8XsHMM+n9Jb8SToacspXwF0QrciVVowfdN/tyyP9ZVYzM9jSTm7K0NA
upfUccDxhgSNv6zNQNKix2qoNIQQuq4OxQpaqXUCgYEAxzmgT/D+wrL+YwtAbqQf
iaePmVV/dy+T98R+5DGtDOtY74WT4IWkLK40ox+h8sNVMUplhhOvQoiqDk4uv+0C
7E+CWuJRZ90OVFXf0kMr80DLqyAT/VI5aObkXzeSF+EfOGnNyH9ljnPuiiHq3dgu
sFut1oMLg7j/6IWg710ETNkCgYEAz/PWMHU1rUeMzw3g5mqBQdNSQErk5Q5sioNM
uNj1O7BGkU03LtYuqiF0n1QjhWo2Lquz8Azmblti/uVfLMQqPAJRgR0ztFvaljE8
aScorJ1w+7j5IU7FRriZBrmFsWslI+nLKPa0xIGaWLzLS2PFjnzgyToEBBO61dzr
tqgHuLECgYBA5nJsbMc36x55d61/4XLkX4Pmoh/LtTEZt9LyoWXUOq86vZuP1W/u
fcCY7mv+lRAoiAC1Z38YjGgJmZINH9EoaVnZJNeaO/86qemnlwYQ5/DIViycPZWI
oXO3sikOr2yrDS95jHjVzU0iW3xzu8bM9D01swBx0T5kYKWZo4ywpQ==
-----END RSA PRIVATE KEY-----
"""

FIELD_FORMATS = [
    "discrete",
    "image",
    "date-time",
    "table",
    "ip",
    "subnet",
    "version",
    "password",
]
SCHEMA_FIELD_FORMATS = [
    "image",
    "date-time",
    "table",
    "logo",
    "tag",
    "ip",
    "subnet",
    "version",
]
SCHEMA_TYPES = ["string", "bool", "array", "integer", "number", "file"]
TAGS = ["badwolf_tag_1", "badwolf_tag_2"]

LINUX_QUERY = 'specific_data.data.os.type == "Linux"'
SHELL_ACTION_NAME = "Badwolf Shell Action"
SHELL_ACTION_CMD = "echo 'Badwolf' > /tmp/badwolf.txt"
DEPLOY_ACTION_NAME = "Badwolf Deploy Action"
DEPLOY_FILE_NAME = "badwolf.sh"
DEPLOY_FILE_CONTENTS = b"#!/bin/bash\necho badwolf!"

CREATE_EC_NAME = "Badwolf EC Example"
CREATE_EC_TRIGGER1 = {
    "name": "Trigger",
    "conditions": {
        "new_entities": False,
        "previous_entities": False,
        "above": 1,
        "below": 0,
    },
    "period": "never",
    "run_on": "AllEntities",
}

CREATE_EC_ACTION_MAIN = {
    "name": "Badwolf Create Notification {}".format(time.time()),
    "action": {"action_name": "create_notification", "config": {}},
}


CSV_FILENAME = "badwolf.csv"
CSV_FIELDS = ["mac_address", "field1"]
CSV_ROW = ["01:37:53:9E:82:7C", "e"]
CSV_FILECONTENTS = [",".join(CSV_FIELDS), ",".join(CSV_ROW)]
CSV_FILECONTENT_STR = "\r\n".join(CSV_FILECONTENTS) + "\r\n"
CSV_FILECONTENT_BYTES = CSV_FILECONTENT_STR.encode()

NORM_TYPES = [
    "string",
    "string_datetime",
    "string_image",
    "string_version",
    "string_ipaddress",
    "bool",
    "integer",
    "number",
    "complex_table",
    "complex",
    "list_integer",
    "list_string",
    "list_string",
    "list_string_version",
    "list_string_datetime",
    "list_string_subnet",
    "list_string",
    "list_string_ipaddress",
    "complex_complex",
]
SCHEMA_STR = {"name": "schema_str", "type": "string", "required": False}
SCHEMA_STR_PASSWORD = {
    "name": "schema_str_password",
    "type": "string",
    "format": "password",
    "required": False,
}
SCHEMA_STR_ENUM = {
    "name": "schema_str_enum",
    "type": "string",
    "enum": ["badwolf"],
    "required": False,
}
SCHEMA_INT = {"name": "schema_int", "type": "integer", "required": False}
SCHEMA_NUM = {"name": "schema_num", "type": "number", "required": False}
SCHEMA_BOOL = {"name": "schema_bool", "type": "bool", "required": False}
SCHEMA_ARRAY = {"name": "schema_array", "type": "array", "required": False}
SCHEMA_FILE = {"name": "schema_file", "type": "file", "required": False}
SCHEMA_UNKNOWN = {"name": "schema_unknown", "type": "badwolf", "required": False}
SCHEMAS = [
    SCHEMA_STR,
    SCHEMA_STR_PASSWORD,
    SCHEMA_STR_ENUM,
    SCHEMA_INT,
    SCHEMA_NUM,
    SCHEMA_BOOL,
    SCHEMA_ARRAY,
    SCHEMA_FILE,
]
SCHEMAS_DICT = {x["name"]: x for x in SCHEMAS}

TEST_USER = "badwolfy"
TEST_ROLE = "badwolfy"
TEST_PERM = "Restricted"
ABOUT_KEYS = ["Build Date", "Version"]
ABOUT_KEYS_EMPTY_OK = ["Version"]
# ABOUT on BUILD: ["Build Date", "Commit Date", "Commit Hash", "Version"]
# ABOUT on RELEASE: ["Build Date", "Customer ID", "Version"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Package setup."""
import codecs
import os

from setuptools import find_packages, setup

PROJECT = "axonius_api_client"
SHELL_CMD = "axonshell"
HERE = os.path.abspath(os.path.dirname(__file__))
VERSION_PATH = os.path.join(HERE, PROJECT, "version.py")


ABOUT = {}
with codecs.open(VERSION_PATH, "r", "utf-8") as fh:
    CONTENTS = "\n".join(a for a in fh.readlines() if not a.startswith("#"))
    exec(CONTENTS, ABOUT)


with codecs.open("README.md", "r", "utf-8") as f:
    README = f.read()


install_requires = [
    "requests[security,socks]>=2.23.0",
    "python-dotenv>=0.12.0",
    "python-dateutil>=2.8.1",
    "click>=7.1.1",
    "pyreadline>=2.1 ; platform_system == 'Windows'",
    "tabulate>=0.8.7",
    "jsonstreams>=0.4.2",
]

setup(
    name=ABOUT["__title__"],
    version=ABOUT["__version__"],
    description=ABOUT["__description__"],
    long_description=README,
    long_description_content_type="text/markdown",
    author=ABOUT["__author__"],
    author_email=ABOUT["__author_email__"],
    url=ABOUT["__url__"],
    packages=find_packages(),
    package_data={"": ["LICENSE"]},
    include_package_data=True,
    python_requires=">=3.5",
    install_requires=install_requires,
    keywords=["Axonius", "API Library"],
    tests_require=["pytest", "pytest-cov", "pytest-httpbin", "coverage"],
    license=ABOUT["__license__"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    entry_points={"console_scripts": [f"{SHELL_CMD}={PROJECT}.cli:cli"]},
)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ....exceptions import NotFoundError
from ....tools import json_dump
from ...context import CONTEXT_SETTINGS, click
from ...options import AUTH, INPUT_FILE, add_options

OPTIONS = [*AUTH, INPUT_FILE]


@click.command(name="add-from-json", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, input_file, **kwargs):
    """Add a saved query from a JSON file."""
    new_sq = ctx.obj.read_stream_json(stream=input_file, expect=dict)

    remove_keys = [
        "date_fetched",
        "predefined",
        "timestamp",
        "updated_by",
        "user_id",
        "uuid",
    ]
    [new_sq.pop(x, None) for x in remove_keys]

    set_defaults = {"query_type": "saved"}
    new_sq.update(set_defaults)

    ensure_keys = ["name", "view"]
    missing_keys = [x for x in ensure_keys if x not in new_sq]

    if missing_keys:
        missing_keys = ", ".join(missing_keys)
        ctx.obj.echo_error(f"Missing required keys: {missing_keys}")

    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    p_grp = ctx.parent.parent.command.name
    apiobj = getattr(client, p_grp)

    sq_name = new_sq["name"]

    try:
        row = apiobj.saved_query.get_by_name(value=sq_name)
        ctx.obj.echo_error(f"Saved query {sq_name!r} already exists, can not add")
    except NotFoundError:
        ctx.obj.echo_ok(f"Saved query {sq_name!r} does not exist, will add")

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        sq_uuid = apiobj.saved_query._add(data=new_sq)
        row = apiobj.saved_query.get_by_uuid(value=sq_uuid)

    ctx.obj.echo_ok(f"Successfully created saved query: {row['name']}")

    click.secho(json_dump(row))

    ctx.exit(0)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
HELPSTR_AUTH = """
Detailed help for authentication:

--url: URL of an Axonius instance

    This can be an IP address, a hostname, or a fully qualified url (and can
    optionally include the port as :port)

--key: API Key of user in an Axonius instance

    See https://axonius-api-client.readthedocs.io/en/latest/main/axtokens.html

--secret: API Secret of user in an Axonius instance

    See https://axonius-api-client.readthedocs.io/en/latest/main/axtokens.html
"""


HELPSTR_EXPORT_ASSET = """
Detailed help for exporting assets:

--export-file:

    File to send data to

    If not supplied data will go to STDOUT, which can be redirected using
    standard OS redirection

    If supplied as a non-absolute path, it will be created as relative
    to --export-path

    If the file does not exist, it will get created with read/write permissions
    for the current user only

--export-path:

    Directory to write --export-file to

    If the directory does not exist, it will get created with read/write
    permissions for the current user only

    The default is the current working directory

--export-table-format:

    Table format to use for export-format=table

    For --export-format=json or csv: ignored

--export-table-max-rows:

    Only return this many rows for export-format=table

    The table format is not very useful for returning large datasets, so we limit
    this to a very low number by default

    For --export-format=json or csv: ignored

--export-include-schema/--no-export-include-schema:

    Add schema information to the export

    For --export-format=json: Add a row at the end of the list of assets with
    the full schema of all fields returned for each asset

    For --export-format=csv: Add 2 rows below the header row with the internal
    field name and the normalized value type

    For --export-format=table: ignored

--export-use-titles/--no-export-use-titles:

    Rename fields from internal field names to their column titles

    For --export-format=json: Rename the keys for each row from the internal
    field name to the field title

    For --export-format=csv or table: Use the field titles as headers
    instead of the internal field name (built-in behavior, can not turn off)

--export-join/--no-export-join:

    Join multivalue fields using --export-join-value

    For --export-format=csv or table: built-in behavior, can not turn off

--export-join-trim:

    Character length to trim joined multivalue fields

    When --export-join=True: if joined value is over this many characters,
    trim it and add a string showing why it was trimmed

    The default is just below the maximum cell character length of Excel (32767)

    If 0 is supplied, no trimming will be applied

--export-explode:

    Flatten and explode a fields values into multiple rows

    This will create new rows for each item in the supplied fields value
    with all other fields duplicated in each new row

    Missing complex sub-fields will be added with a value set to
    --field-null-value.

    This must be a field that has been supplied to --field and the
    field name can be:

    - a fully qualified field name: specific_data.data.installed_software

    - a short field name: installed_software

--export-flatten/--no-export-flatten:

    Remove complex fields and re-add their sub-field values to the row

    For example, open_ports has 3 sub-fields: port_id, protocol, service_name

    Flattening open_ports would remove it from each asset, and add
    open_ports.port_id, open_ports.protocol, and open_ports.service_name
    with their values

    Sub-fields that are not returned will be set using --export-null-value

--export-field-exclude:

    Fields to exclude from each row

    Supply multiple --export-field-exclude for each value

    This must be a field that has been supplied to --field and the field
    name can be any one of:

    - fully qualified field name: specific_data.data.installed_software

    - short field name: installed_software

    - fully qualified sub-field name: specific_data.data.installed_software.name

    - short sub-field name: installed_software.name

    - sub-field name: name
"""

HELPSTR_QUERY = """
Detailed help for supplying a query:

--query:
    Query built from the Query wizard in the GUI

    All assets are returned if this is not supplied

--query-file:
    Path to a file to override --query

    Path to a file containing a query built from the Query wizard in the GUI.
    Overrides --query if both are supplied

    This can be used to avoid dealing with quote escaping issues in varying
    command lines

    Supply '-' to read the file from STDIN.

"""

HELPSTR_SELECT_FIELDS = """
Detailed help for selecting fields:

--fields:
    Fields to include in the format of adapter:field

    Can supply multiple --field values.

    Examples:

    - hostname - get the aggregated hostname field
      (same as: generic:hostname, aggregated:hostname, specific_data.data.hostname)

    - aws:hostname - get the AWS adapter specific hostname field
      (same as adapters_data.aws_adapter)

    - hostname,network_interfaces - get the aggregated hostname and
      network_interfaces fields
"""


HELPSTRS = {}
HELPSTRS["auth"] = HELPSTR_AUTH
HELPSTRS["assetexport"] = HELPSTR_EXPORT_ASSET
HELPSTRS["selectfields"] = HELPSTR_SELECT_FIELDS
HELPSTRS["query"] = HELPSTR_QUERY



# -*- coding: utf-8 -*-
"""Test suite."""
import copy

import pytest

from axonius_api_client.api.parsers.config import (
    config_check,
    config_check_array,
    config_check_bool,
    config_check_file,
    config_check_int,
    config_check_str,
    config_default,
    config_empty,
    config_required,
    config_unchanged,
    config_unknown,
    is_uploaded_file,
    parse_schema,
)
from axonius_api_client.constants import SETTING_UNCHANGED
from axonius_api_client.exceptions import (
    ApiError,
    ConfigInvalidValue,
    ConfigRequired,
    ConfigUnchanged,
    ConfigUnknown,
)

from ...meta import (
    SCHEMA_ARRAY,
    SCHEMA_BOOL,
    SCHEMA_FILE,
    SCHEMA_INT,
    SCHEMA_NUM,
    SCHEMA_STR,
    SCHEMA_STR_ENUM,
    SCHEMA_STR_PASSWORD,
    SCHEMA_UNKNOWN,
    SCHEMAS,
    SCHEMAS_DICT,
)


def test_config_check_str():
    """Pass."""
    value = "test"
    result = config_check_str(value=value, schema=SCHEMA_STR, source="badwolf")
    assert result == value

    result = config_check(value=value, schema=SCHEMA_STR, source="badwolf")
    assert result == value


def test_config_check_str_bad():
    """Pass."""
    value = {"11111": "x"}
    with pytest.raises(ConfigInvalidValue):
        config_check_str(value=value, schema=SCHEMA_STR, source="badwolf")

    with pytest.raises(ConfigInvalidValue):
        config_check(value=value, schema=SCHEMA_STR, source="badwolf")


@pytest.mark.parametrize(
    "value",
    [
        SETTING_UNCHANGED,
        SETTING_UNCHANGED[0],
        str(SETTING_UNCHANGED),
        str(SETTING_UNCHANGED[0]),
    ],
)
def test_config_check_str_password_unchanged(value):
    """Pass."""
    result = config_check_str(value=value, schema=SCHEMA_STR_PASSWORD, source="badwolf")
    assert result == SETTING_UNCHANGED

    result = config_check(value=value, schema=SCHEMA_STR_PASSWORD, source="badwolf")
    assert result == SETTING_UNCHANGED


def test_config_check_str_password():
    """Pass."""
    value = "badwolf"
    result = config_check_str(value=value, schema=SCHEMA_STR_PASSWORD, source="badwolf")
    assert result == value

    result = config_check(value=value, schema=SCHEMA_STR_PASSWORD, source="badwolf")
    assert result == value


def test_config_check_str_enum():
    """Pass."""
    value = "badwolf"
    result = config_check_str(value=value, schema=SCHEMA_STR_ENUM, source="badwolf")
    assert result == value

    result = config_check(value=value, schema=SCHEMA_STR_ENUM, source="badwolf")
    assert result == value


def test_config_check_str_enum_bad():
    """Pass."""
    value = "xxxx"
    with pytest.raises(ConfigInvalidValue):
        config_check_str(value=value, schema=SCHEMA_STR_ENUM, source="badwolf")

    with pytest.raises(ConfigInvalidValue):
        config_check(value=value, schema=SCHEMA_STR_ENUM, source="badwolf")


@pytest.mark.parametrize("schema", [SCHEMA_INT, SCHEMA_NUM])
def test_config_check_int(schema):
    """Pass."""
    value = 111
    result = config_check_int(value=value, schema=schema, source="badwolf")
    assert result == value

    result = config_check(value=value, schema=schema, source="badwolf")
    assert result == value


@pytest.mark.parametrize("schema", [SCHEMA_INT, SCHEMA_NUM])
def test_config_check_int_str(schema):
    """Pass."""
    value = "111"
    exp = 111
    result = config_check_int(value=value, schema=schema, source="badwolf")
    assert result == exp

    result = config_check(value=value, schema=schema, source="badwolf")
    assert result == exp


@pytest.mark.parametrize("schema", [SCHEMA_INT, SCHEMA_NUM])
def test_config_check_int_bad(schema):
    """Pass."""
    value = "badwolf"
    with pytest.raises(ConfigInvalidValue):
        config_check_int(value=value, schema=schema, source="badwolf")
    with pytest.raises(ConfigInvalidValue):
        config_check(value=value, schema=schema, source="badwolf")


def test_config_check_bool_bad():
    """Pass."""
    value = "badwolf"
    with pytest.raises(ConfigInvalidValue):
        config_check_bool(value=value, schema=SCHEMA_BOOL, source="badwolf")

    with pytest.raises(ConfigInvalidValue):
        config_check(value=value, schema=SCHEMA_BOOL, source="badwolf")


def test_config_check_bool_yes():
    """Pass."""
    value = "yes"
    exp = True
    result = config_check_bool(value=value, schema=SCHEMA_BOOL, source="badwolf")
    assert result == exp

    result = config_check(value=value, schema=SCHEMA_BOOL, source="badwolf")
    assert result == exp


def test_config_check_bool_no():
    """Pass."""
    value = "no"
    exp = False
    result = config_check_bool(value=value, schema=SCHEMA_BOOL, source="badwolf")
    assert result == exp

    result = config_check(value=value, schema=SCHEMA_BOOL, source="badwolf")
    assert result == exp


def test_config_check_array_list():
    """Pass."""
    value = ["badwolf1", "badwolf2"]
    result = config_check_array(value=value, schema=SCHEMA_ARRAY, source="badwolf")
    assert result == value

    result = config_check(value=value, schema=SCHEMA_ARRAY, source="badwolf")
    assert result == value


def test_config_check_array_str():
    """Pass."""
    value = "badwolf1, badwolf2"
    exp = ["badwolf1", "badwolf2"]
    result = config_check_array(value=value, schema=SCHEMA_ARRAY, source="badwolf")
    assert result == exp

    result = config_check(value=value, schema=SCHEMA_ARRAY, source="badwolf")
    assert result == exp


@pytest.mark.parametrize("value", [1111, [111]])
def test_config_check_array_bad(value):
    """Pass."""
    with pytest.raises(ConfigInvalidValue):
        config_check_array(value=value, schema=SCHEMA_ARRAY, source="badwolf")
    with pytest.raises(ConfigInvalidValue):
        config_check(value=value, schema=SCHEMA_ARRAY, source="badwolf")


def test_config_check_file_uploaded():
    """Pass."""
    value = {"uuid": "x", "filename": "x"}
    result = config_check_file(value=value, schema=SCHEMA_FILE, source="badwolf")
    assert value == result

    result = config_check(value=value, schema=SCHEMA_FILE, source="badwolf")
    assert value == result


def test_config_check_file_callback():
    """Pass."""

    def mock_callback(value, schema, callbacks, source):
        return {"uuid": "x", "filename": "x"}

    value = "badwolf"
    exp = {"uuid": "x", "filename": "x"}
    callbacks = {"cb_file": mock_callback}
    result = config_check_file(
        value=value, schema=SCHEMA_FILE, source="badwolf", callbacks=callbacks
    )
    assert result == exp

    result = config_check(
        value=value, schema=SCHEMA_FILE, source="badwolf", callbacks=callbacks
    )
    assert result == exp


def test_config_check_file_unsupported():
    """Pass."""
    value = "x"
    with pytest.raises(ApiError):
        config_check_file(value=value, schema=SCHEMA_FILE, source="badwolf")

    with pytest.raises(ApiError):
        config_check(value=value, schema=SCHEMA_FILE, source="badwolf")


def test_config_check_unknown_schema():
    """Pass."""
    value = "badwolf"
    with pytest.raises(ApiError):
        config_check(value=value, schema=SCHEMA_UNKNOWN, source="badwolf")


@pytest.mark.parametrize("schema", SCHEMAS)
def test_config_check_none_ok(schema):
    """Pass."""
    value = None
    result = config_check(value=value, schema=schema, none_ok=True, source="badwolf")
    assert result == value


@pytest.mark.parametrize("schema", SCHEMAS)
def test_config_check_none_bad(schema):
    """Pass."""
    value = None
    with pytest.raises(ConfigInvalidValue):
        config_check(value=value, schema=schema, none_ok=False, source="badwolf")


def test_parse_schema_req_true():
    """Pass."""
    schema = {"items": [{"name": "x"}], "required": ["x"]}
    exp = {"x": {"name": "x", "required": True}}
    result = parse_schema(raw=schema)
    assert result == exp


def test_parse_schema_req_false():
    """Pass."""
    schema = {"items": [{"name": "x"}], "required": []}
    exp = {"x": {"name": "x", "required": False}}
    result = parse_schema(raw=schema)
    assert result == exp


def test_is_uploaded_file_true():
    """Pass."""
    value = {"uuid": "x", "filename": "x"}
    exp = True, value
    result = is_uploaded_file(value=value)
    assert result == exp


def test_is_uploaded_file_true_json():
    """Pass."""
    value = '{"uuid": "x", "filename": "x"}'
    exp = True, {"uuid": "x", "filename": "x"}
    result = is_uploaded_file(value=value)
    assert result == exp


def test_is_uploaded_file_false_json():
    """Pass."""
    value = '{"uu": "x", "filename": "x"}'
    exp = False, value
    result = is_uploaded_file(value=value)
    assert result == exp


@pytest.mark.parametrize(
    "value",
    [
        "x",
        {},
        {"uuid": "x"},
        {"uuid": ""},
        {"filename": "x"},
        {"filename": ""},
        {"uuid": "", "filename": ""},
        {"uuid": [""], "filename": [""]},
    ],
)
def test_is_uploaded_file_false(value):
    """Pass."""
    exp = False, value
    result = is_uploaded_file(value=value)
    assert result == exp


def test_config_empty_ok():
    """Pass."""
    config_empty(
        schemas=SCHEMAS_DICT, new_config={"badwolf": "badwolf"}, source="badwolf"
    )


def test_config_empty_bad():
    """Pass."""
    with pytest.raises(ConfigRequired):
        config_empty(schemas=SCHEMAS_DICT, new_config={}, source="badwolf")


@pytest.mark.parametrize(
    "value", [{}, {"x": "x"}, {SCHEMA_STR["name"]: "vvv"}],
)
def test_config_required_ok(value):
    """Pass."""
    schemas = copy.deepcopy(SCHEMAS_DICT)
    for name, schema in schemas.items():
        schema["required"] = False
    config_required(schemas=schemas, new_config=value, source="badwolf")


def test_config_required_bad():
    """Pass."""
    schemas = copy.deepcopy(SCHEMAS_DICT)
    for name, schema in schemas.items():
        schema["required"] = True
    with pytest.raises(ConfigRequired):
        config_required(schemas=schemas, new_config={}, source="badwolf")


def test_config_required_ignored():
    """Pass."""
    schemas = copy.deepcopy(SCHEMAS_DICT)
    for name, schema in schemas.items():
        schema["required"] = True
    ignores = list(schemas)
    config_required(schemas=schemas, new_config={}, source="badwolf", ignores=ignores)


@pytest.mark.parametrize(
    "value",
    [{}, {"x": "x"}, {SCHEMA_STR["name"]: "moo"}, {SCHEMA_STR["name"]: "moo", "x": "x"}],
)
def test_config_default(value):
    """Pass."""
    schemas = copy.deepcopy(SCHEMAS_DICT)
    exp = {}
    for name, schema in schemas.items():
        schema["default"] = "badwolf"
        exp[name] = "badwolf"
    exp.update(value)
    result = config_default(
        schemas=schemas, new_config=copy.deepcopy(value), source="badwolf"
    )
    assert result == exp


def test_config_changed_ok():
    """Pass."""
    config_unchanged(
        schemas=SCHEMAS_DICT,
        old_config={"x": "x"},
        new_config={"x": "y"},
        source="badwolf",
    )


@pytest.mark.parametrize(
    "value", [{}, {"x": "x"}],
)
def test_config_changed_bad(value):
    """Pass."""
    with pytest.raises(ConfigUnchanged):
        config_unchanged(
            schemas=SCHEMAS_DICT,
            old_config={"x": "x"},
            new_config=value,
            source="badwolf",
        )


def test_config_unknown_ok():
    """Pass."""
    schema_name = SCHEMA_STR["name"]
    config = {schema_name: "badwolf"}
    config_unknown(schemas=SCHEMAS_DICT, new_config=config, source="badwolf")


def test_config_unknown_bad():
    """Pass."""
    config = {"badwolf": "badwolf"}
    with pytest.raises(ConfigUnknown):
        config_unknown(schemas=SCHEMAS_DICT, new_config=config, source="badwolf")



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import atexit
import os

import click

import axonius_api_client as axonapi

from ...constants import PY36
from ...tools import echo_error, json_reload, pathlib
from ..context import CONTEXT_SETTINGS
from ..options import AUTH, add_options

SHELL_BANNER = """Welcome human. We have some refreshments available for you:

    - ctx: Click context object
    - client/c: Instance of axonius_api_client.connect.Connect
    - jdump/j: Helper function to pretty print python objects
    - axonapi: API client package itself

API Objects:
    - adapters/a: Instance of axonius_api_client.api.Adapters
    - devices/d: Instance of axonius_api_client.api.Devices
    - users/u: Instance of axonius_api_client.api.Users
    - system/s: Instance of axonius_api_client.api.System
"""

SHELL_EXIT = """Goodbye human. We hope you enjoyed your stay."""

HISTPATH = os.path.expanduser("~")

HISTFILE = ".python_history"


def jdump(data):
    """Pass."""
    print(json_reload(data))


@click.command(name="shell", context_settings=CONTEXT_SETTINGS)
@add_options(AUTH)
@click.pass_context
def cmd(ctx, url, key, secret):  # noqa: D301
    """Start an interactive python shell.

    The shell will authenticate to Axonius, setup autocompletion, enable history,
    and create the following objects:

    \b
        - ctx: Click context object
        - client/c: Instance of axonius_api_client.connect.Connect
        - devices/d: Instance of axonius_api_client.api.Devices
        - users/u: Instance of axonius_api_client.api.Users
        - adapters/a: Instance of axonius_api_client.api.Adapters
        - system/s: Instance of axonius_api_client.api.System
        - axonapi: API Client package itself
        - jdump/j: Helper function to pretty print python objects

    """
    client = ctx.obj.start_client(url=url, key=key, secret=secret, save_history=True)

    client._http.save_history = True

    shellvars = {
        "ctx": ctx,
        "client": client,
        "c": client,
        "devices": client.devices,
        "d": client.devices,
        "users": client.users,
        "u": client.users,
        "adapters": client.adapters,
        "a": client.adapters,
        "jdump": jdump,
        "j": jdump,
        "s": client.system,
        "system": client.system,
        "axonapi": axonapi,
    }

    spawn_shell(shellvars)


def write_hist_file():
    """Pass."""
    import readline

    histpath = pathlib.Path(HISTPATH)
    histfile = histpath / HISTFILE

    histpath.mkdir(mode=0o700, exist_ok=True)
    histfile.touch(mode=0o600, exist_ok=True)

    readline.write_history_file(format(histfile))


def register_readline(shellvars=None):
    """Pass."""
    try:
        import readline
    except Exception:  # pragma: no cover
        import pyreadline as readline

    import rlcompleter

    shellvars = shellvars or {}

    histpath = pathlib.Path(HISTPATH)
    histfile = histpath / HISTFILE

    histpath.mkdir(mode=0o700, exist_ok=True)
    histfile.touch(mode=0o600, exist_ok=True)

    try:
        readline.read_history_file(format(histfile))
        atexit.register(write_hist_file)

        readline.set_completer(rlcompleter.Completer(shellvars).complete)

        readline_doc = getattr(readline, "__doc__", "")
        is_libedit = readline_doc and "libedit" in readline_doc

        pab = "bind ^I rl_complete" if is_libedit else "tab: complete"
        readline.parse_and_bind(pab)
    except Exception as exc:
        msg = "Unable to register history and autocomplete:\n{}".format(exc)
        echo_error(msg, abort=False)


def spawn_shell(shellvars=None):
    """Pass."""
    import code

    shellvars = shellvars or {}
    register_readline(shellvars)

    args = {"local": shellvars, "banner": SHELL_BANNER}

    if PY36:
        args["exitmsg"] = SHELL_EXIT

    code.interact(**args)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
# from ...api.parsers.tables import tablize_adapters
from ...tools import json_dump
from ..context import click

CONFIG_EXPORT = click.option(
    "--export-format",
    "-xf",
    "export_format",
    type=click.Choice(["json-full", "json"]),
    help="Format of to export data in",
    default="json",
    show_envvar=True,
    show_default=True,
)

CONFIG_TYPE = click.option(
    "--generic/--specific",
    "-g/-s",
    "generic",
    default=True,
    show_envvar=True,
    show_default=True,
    help="Generic or adapter specific config",
)


def config_export(ctx, rows, export_format):
    """Pass."""
    if export_format == "json":
        click.secho(json_dump(rows["config"]))
    elif export_format == "json-full":
        click.secho(json_dump(rows))
    ctx.exit(0)



# -*- coding: utf-8 -*-
"""Test suite."""
import pytest

from axonius_api_client.api.parsers.fields import get_type_norm
from axonius_api_client.exceptions import ApiError


def test_get_type_norm_error():
    """Pass."""
    field = {"type": "badwolf", "name": "badwolf"}
    with pytest.raises(ApiError):
        get_type_norm(field=field)



# -*- coding: utf-8 -*-
"""API models for working with device and user assets."""
import copy
import sys

from ...constants import (DEFAULT_PATH, FIELD_JOINER, FIELD_TRIM_LEN,
                          FIELD_TRIM_STR, GET_SCHEMA_KEYS, SCHEMAS_CUSTOM)
from ...exceptions import ApiError
from ...tools import echo_error, echo_ok, get_path, listify


def custom_schemas(**kwargs):
    """Pass."""
    schemas = []
    if kwargs.get("report_adapters_missing", False):
        schemas += list(SCHEMAS_CUSTOM["report_adapters_missing"].values())
    return schemas


class Base:
    """Object for handling callbacks for assets."""

    CB_NAME = "base"

    def __init__(
        self, apiobj, fields, query=None, state=None, fields_map=None, getargs=None
    ):
        """Object for handling callbacks for assets."""
        self.LOG = apiobj.LOG.getChild(self.__class__.__name__)
        """:obj:`logging.Logger`: logger for this object."""

        self.APIOBJ = apiobj
        """:obj:`AssetMixin`: assets object."""

        self.QUERY = query
        """:obj:`str`: query supplied to get assets method."""

        self.STATE = state or {}
        """:obj:`dict`: state dict used by get assets method to track paging."""

        self.GETARGS = getargs or {}
        """:obj:`dict`: original kwargs supplied to get assets method."""

        self.RAN = []
        """:obj:`list` of :obj:`str`: used by callbacks to see if they've run already."""

        self.TAG_ROWS_ADD = []
        """:obj:`list` of :obj:`dict`: assets to add tags to in do_tagging."""

        self.TAG_ROWS_REMOVE = []
        """:obj:`list` of :obj:`dict`: assets to remove tags from in do_tagging."""

        self.FIELDS = [x for x in apiobj.FIELDS_API if x not in fields] + fields
        """:obj:`list` of :obj:`str`: field names selected originally."""

        self.FIELDS_MAP = fields_map or apiobj.fields.get()
        """:obj:`dict`: map of adapter -> field schemas."""

        self.SCHEMAS_CUSTOM = custom_schemas(**self.GETARGS)
        """:obj:`list` of :obj:`dict`: schemas of custom fields."""

        self.SCHEMAS_ALL = [y for x in self.FIELDS_MAP.values() for y in x]
        """:obj:`list` of :obj:`dict`: field schemas for all adapters."""

        self.SCHEMAS_FIELDS = [self.get_schema(value=x) for x in self.FIELDS]
        """:obj:`list` of :obj:`dict`: schemas of fields selected originally."""

    def get_schema(self, value, schemas=None, qual_only=True, **kwargs):
        """Pass."""
        if qual_only:
            kwargs["keys"] = ["name_qual"]
        schemas = schemas or (self.SCHEMAS_ALL + self.SCHEMAS_CUSTOM)
        return self.APIOBJ.fields.get_field_schema(
            value=value, schemas=schemas, **kwargs
        )

    def schemas_final_complex(self, schema, final):
        """Pass."""
        sub_schemas = self.filter_sub_schemas(schema=schema)
        final.update({x["name_qual"]: x for x in sub_schemas})

    def schemas_final_simple(self, schema, final):
        """Pass."""
        final.setdefault(schema["name_qual"], schema)

    def do_schemas_final(self, schema, explode, flat, final):
        """Pass."""
        if schema["is_complex"] and (schema["name_qual"] == explode or flat):
            self.schemas_final_complex(schema=schema, final=final)
        else:
            self.schemas_final_simple(schema=schema, final=final)

    def schemas_final(self, flat=False):
        """Predict the future schemas that will be returned."""
        flat = True if self.GETARGS.get("field_flatten", False) else flat

        schemas = self.filter_schemas(schemas=self.SCHEMAS_FIELDS)

        explode = self.explode_schema.get("name_qual", "")

        final = {}

        for schema in schemas:
            self.do_schemas_final(schema=schema, explode=explode, flat=flat, final=final)

        final = list(final.values())
        final += self.SCHEMAS_CUSTOM
        return final

    def start(self, **kwargs):
        """Run start callbacks."""
        join = "\n   - "
        self.echo(msg=f"Starting {self}")

        cbargs = ["{}={!r}".format(k, v) for k, v in self.GETARGS.items()]
        cbargs = "\n  " + "\n  ".join(cbargs)
        self.LOG.debug(f"Callback arguments: {cbargs}")

        config = join + join.join(self.args_strs)
        self.echo(msg=f"Configuration: {config}")
        self.echo(msg=f"Query: {self.QUERY}")

        schemas = self.SCHEMAS_FIELDS
        schemas_pretty = self.APIOBJ.fields._prettify_schemas(schemas=schemas)
        schemas_pretty = join + join.join(schemas_pretty)
        self.echo(msg=f"Fields: {schemas_pretty}")

    def stop(self, **kwargs):
        """Run stop callbacks."""
        self.do_tagging()
        self.echo(msg=f"Stopping {self}")

    def row(self, row):
        """Handle callbacks for an asset."""
        self.first_page()
        row = self.tags_add(row=row)
        row = self.tags_remove(row=row)
        row = self.report_adapters_missing(row=row)
        row = self.field_null(row=row)
        row = self.field_excludes(row=row)
        rows = self.field_explode(row=row)
        rows = [self.field_flatten(row=x) for x in rows]
        rows = [self.field_join(row=x) for x in rows]
        rows = [self.field_titles(row=x) for x in rows]
        return rows

    def field_null(self, row):
        """Asset callback to add None values to fields not returned."""
        if not self.GETARGS.get("field_null", False):
            return row

        for schema in self.SCHEMAS_FIELDS:
            self.do_null(field=schema["name_qual"], row=row, schema=schema)
        return row

    def null_complex(self, row, field, schema):
        """Pass."""
        row[field] = listify(row.get(field, []))
        for entry in row[field]:
            sub_schemas = self.filter_sub_schemas(schema=schema)
            for sub_schema in sub_schemas:
                self.do_null(field=sub_schema["name"], schema=sub_schema, row=entry)

    def null_simple(self, row, field, schema):
        """Pass."""
        null_value = self.GETARGS.get("field_null_value", None)
        row[field] = row.get(field, null_value)

    def do_null(self, row, field, schema):
        """Null out missing fields."""
        if self.is_excluded(schema=schema):
            return

        if schema["is_complex"]:
            self.null_complex(row=row, field=field, schema=schema)
        else:
            self.null_simple(row=row, field=field, schema=schema)

    def exclude_complex_sub(self, row, field, items, schema):
        """Pass."""
        if self.is_excluded(schema=schema):
            for item in items:
                item.pop(schema["name"], None)

    def exclude_complex(self, row, field, schema):
        """Pass."""
        if schema["is_complex"]:
            sub_schemas = schema["sub_fields"]
            items = listify(row.get(field, []))
            for sub_schema in sub_schemas:
                self.exclude_complex_sub(
                    row=row, field=field, items=items, schema=sub_schema
                )

    def exclude_simple(self, row, field, schema):
        """Pass."""
        if not schema["is_complex"] and self.is_excluded(schema=schema):
            row.pop(field, None)

    def do_exclude(self, row, field, schema):
        """Pass."""
        self.exclude_simple(row=row, field=field, schema=schema)
        self.exclude_complex(row=row, field=field, schema=schema)

    def field_excludes(self, row):
        """Asset callback to remove fields from row."""
        if not self.GETARGS.get("field_excludes", []):
            return row

        for field in list(row):
            schema = self.get_schema(value=field)
            self.do_exclude(row=row, field=field, schema=schema)

        return row

    def field_join(self, row):
        """Join values."""
        if not self.GETARGS.get("field_join", False):
            return row

        joiner = self.GETARGS.get("field_join_value", FIELD_JOINER)
        trim_len = self.GETARGS.get("field_join_trim", FIELD_TRIM_LEN)
        trim_str = FIELD_TRIM_STR

        for field in row:
            if isinstance(row[field], list):
                row[field] = joiner.join([str(x) for x in row[field]])

            if trim_len and isinstance(row[field], str):
                field_len = len(row[field])
                if len(row[field]) >= trim_len:
                    msg = trim_str.format(field_len=field_len, trim_len=trim_len)
                    row[field] = joiner.join([row[field][:trim_len], msg])
        return row

    def field_titles(self, row):
        """Asset callback to change qual name to title."""
        if not self.GETARGS.get("field_titles", False):
            return row

        for field in list(row):
            schema = self.get_schema(value=field)
            row[schema["column_title"]] = row.pop(field)
        return row

    def flatten_complex_sub(self, row, field, items, schema):
        """Asset callback to flatten complex fields."""
        null_value = self.GETARGS.get("field_null_value", None)
        sub_name = schema["name"]
        sub_name_qual = schema["name_qual"]
        row[sub_name_qual] = []
        for item in items:
            value = item.pop(sub_name, null_value)
            value = value if isinstance(value, list) else [value]
            row[sub_name_qual] += value

    def flatten_complex(self, row, field, schema):
        """Asset callback to flatten complex fields."""
        if schema["is_complex"] and not self.is_excluded(schema=schema):
            items = listify(row.pop(field, []))
            sub_schemas = self.filter_sub_schemas(schema=schema)
            for sub_schema in sub_schemas:
                self.flatten_complex_sub(
                    row=row, field=field, items=items, schema=sub_schema
                )

    def field_flatten(self, row):
        """Asset callback to flatten complex fields."""
        if not self.GETARGS.get("field_flatten", False):
            return row

        for field in list(row):
            schema = self.get_schema(value=field)
            self.flatten_complex(row=row, field=field, schema=schema)
        return row

    def explode_complex_sub(self, row, rows, items, schema):
        """Pass."""
        null_value = self.GETARGS.get("field_null_value", None)
        sub_name = schema["name"]
        sub_name_qual = schema["name_qual"]

        for idx, item in enumerate(items):
            rows.setdefault(idx, copy.deepcopy(row))
            rows[idx][sub_name_qual] = item.pop(sub_name, null_value)

    def explode_complex(self, row, rows, field, schema):
        """Explode a field into multiple rows."""
        items = listify(row.pop(field))
        sub_schemas = self.filter_sub_schemas(schema=schema)

        for sub_schema in sub_schemas:
            self.explode_complex_sub(row=row, rows=rows, items=items, schema=sub_schema)

    def explode_simple(self, row, rows, field, schema):
        """Explode a field into multiple rows."""
        items = listify(row.pop(field))

        for idx, item in enumerate(items):
            rows.setdefault(idx, copy.deepcopy(row))
            rows[idx][field] = item

    def do_explode(self, row, rows, field, schema):
        """Pass."""
        if schema["is_complex"]:
            self.explode_complex(row=row, rows=rows, field=field, schema=schema)
        else:
            self.explode_simple(row=row, rows=rows, field=field, schema=schema)

    def field_explode(self, row):
        """Explode a field into multiple rows."""
        explode = self.GETARGS.get("field_explode", "")

        if not explode:
            return [row]

        schema = self.explode_schema
        is_excluded = self.is_excluded(schema=schema)
        field = schema["name_qual"]
        if is_excluded or row.get(field, None) in [None, []] or field not in row:
            return [row]

        rows = {}
        self.do_explode(row=row, rows=rows, field=field, schema=schema)
        return [rows[idx] for idx in sorted(rows)]

    @property
    def explode_schema(self):
        """Pass."""
        explode = self.GETARGS.get("field_explode", "")
        if not explode:
            return {}

        valids = []

        for schema in self.SCHEMAS_FIELDS:
            names = [schema[x] for x in GET_SCHEMA_KEYS]
            if explode in names:
                return schema
            valids += names

        msg = f"Explode field {explode!r} not found, valid fields:{valids}"
        self.echo(msg=msg, error=ApiError)

    @property
    def adapter_map(self):
        """Pass."""
        self._adapters_meta = getattr(self, "_adapters_meta", self.APIOBJ.adapters.get())
        amap = {
            "has_cnx": [],
            "all": [],
            "all_fields": [f"{x}_adapter" for x in self.FIELDS_MAP],
        }

        for adapter in self._adapters_meta:
            name_raw = adapter["name_raw"]

            if adapter not in amap["all"]:
                amap["all"].append(name_raw)
            if adapter["cnx"]:
                amap["has_cnx"].append(name_raw)

        amap = {k: list(v) for k, v in amap.items()}
        return amap

    def report_adapters_missing(self, row):
        """Pass."""
        if not self.GETARGS.get("report_adapters_missing", False):
            return row

        schemas = SCHEMAS_CUSTOM["report_adapters_missing"]
        field = schemas["adapters_missing"]["name_qual"]

        adapters_row = row.get("adapters", [])
        adapter_map = self.adapter_map
        missing = []

        for adapter in adapter_map["all"]:
            if adapter in adapters_row:
                continue

            if adapter not in adapter_map["all_fields"]:
                continue

            if adapter not in missing:
                missing.append(adapter)

        row[field] = missing
        return row

    def is_excluded(self, schema):
        """Check if a name supplied to field_excludes matches one of GET_SCHEMA_KEYS."""
        excludes = listify(self.GETARGS.get("field_excludes", []))
        for exclude in excludes:
            for key in GET_SCHEMA_KEYS:
                if schema.get(key, None) == exclude:
                    return True
        return False

    def filter_schemas(self, schemas):
        """Pass."""
        return [x for x in schemas if not self.is_excluded(schema=x)]

    def filter_sub_schemas(self, schema):
        """Pass."""
        schemas = schema["sub_fields"]
        schemas = self.filter_schemas(schemas=schemas)
        schemas = [x for x in schemas if x["is_root"]]
        return schemas

    def tags_add(self, row):
        """Pass."""
        tags = listify(self.GETARGS.get("tags_add", []))
        if not tags:
            return row

        tag_row = {"internal_axon_id": row["internal_axon_id"]}

        if tag_row not in self.TAG_ROWS_ADD:
            self.TAG_ROWS_ADD.append(tag_row)
        return row

    def tags_remove(self, row):
        """Pass."""
        tags = listify(self.GETARGS.get("tags_remove", []))
        if not tags:
            return row

        tag_row = {"internal_axon_id": row["internal_axon_id"]}

        if tag_row not in self.TAG_ROWS_REMOVE:
            self.TAG_ROWS_REMOVE.append(tag_row)

        return row

    def do_tagging(self):
        """Pass."""
        self.do_tag_add()
        self.do_tag_remove()

    def do_tag_add(self):
        """Pass."""
        tags_add = listify(self.GETARGS.get("tags_add", []))
        rows_add = self.TAG_ROWS_ADD
        if tags_add and rows_add:
            self.echo(msg=f"Adding tags {tags_add} to {len(rows_add)} assets")
            self.APIOBJ.labels.add(rows=rows_add, labels=tags_add)

    def do_tag_remove(self):
        """Pass."""
        tags_remove = listify(self.GETARGS.get("tags_remove", []))
        rows_remove = self.TAG_ROWS_REMOVE
        if tags_remove and rows_remove:
            self.echo(msg=f"Removing tags {tags_remove} from {len(rows_remove)} assets")
            self.APIOBJ.labels.remove(rows=rows_remove, labels=tags_remove)

    def first_page(self):
        """Asset callback to echo first page info using an echo method."""
        if "first_page" in self.RAN:
            return

        self.RAN.append("first_page")

        rows_page = self.STATE.get("rows_page")
        rows_total = self.STATE.get("rows_total")
        page_total = self.STATE.get("page_total")

        info = f"{rows_page} rows (total rows {rows_total}, total pages {page_total})"
        self.echo(msg=f"First page received {info}")

    def open_fd_arg(self):
        """Pass."""
        self._fd = self.GETARGS["export_fd"]
        self._fd_close = self.GETARGS.get("export_fd_close", False)
        self.echo(msg=f"Exporting to {self._fd}")
        return self._fd

    def open_fd_path(self):
        """Pass."""
        self._export_file = self.GETARGS.get("export_file", None)
        self._export_path = self.GETARGS.get("export_path", DEFAULT_PATH)
        self._export_overwrite = self.GETARGS.get("export_overwrite", False)

        file_path = get_path(obj=self._export_path)
        file_path.mkdir(mode=0o700, parents=True, exist_ok=True)
        self._file_path = fp = (file_path / self._export_file).resolve()

        if self._file_path.exists():
            self._file_mode = "overwrote"
            mode = "overwriting"
        else:
            self._file_mode = "created"
            mode = "creating"

        if self._file_path.exists() and not self._export_overwrite:
            msg = f"Export file '{fp}' already exists and overwite is False!"
            self.echo(msg=msg, error=ApiError, level="error")

        self._file_path.touch(mode=0o600)
        self._fd_close = self.GETARGS.get("export_fd_close", True)
        self._fd = self._file_path.open(mode="w", encoding="utf-8")
        self.echo(msg=f"Exporting to file '{fp}' ({mode})")
        return self._fd

    def open_fd_stdout(self):
        """Pass."""
        self._file_path = None
        self._fd_close = False
        self._fd = sys.stdout
        self.echo(msg=f"Exporting to stdout")
        return self._fd

    def open_fd(self):
        """Open a file descriptor."""
        if "export_fd" in self.GETARGS:
            self.open_fd_arg()
        elif self.GETARGS.get("export_file", None):
            self.open_fd_path()
        else:
            self.open_fd_stdout()
        return self._fd

    def close_fd(self):
        """Close a file descriptor."""
        self._fd.write("\n")
        if getattr(self, "_fd_close", False):
            self.echo(msg=f"Closing {self._fd}")
            self._fd.close()

    def echo(self, msg, error=None, level="info", level_error="error"):
        """Pass."""
        if self.GETARGS.get("do_echo", False):
            echo_error(msg) if error else echo_ok(msg)
        else:
            if error:
                getattr(self.LOG, level_error)(msg)
                raise error(msg)
            getattr(self.LOG, level)(msg)

    @property
    def args_map(self):
        """Pass."""
        return [
            ["field_excludes", "Exclude fields:", []],
            ["field_flatten", "Flatten complex fields:", False],
            ["field_explode", "Explode field:", None],
            ["field_titles", "Rename fields to titles:", False],
            ["field_join", "Join field values:", False],
            ["field_join_value", "Join field values using:", FIELD_JOINER],
            ["field_join_trim", "Join field character limit:", FIELD_TRIM_LEN],
            ["field_null", "Add missing fields:", False],
            ["field_null_value", "Missing field value:", None],
            ["tags_add", "Add tags:", []],
            ["tags_remove", "Remove tags:", []],
            ["report_adapters_missing", "Report Missing Adapters:", False],
            ["export_file", "Export to file:", None],
            ["export_path", "Export file to path:", DEFAULT_PATH],
            ["export_overwrite", "Export overwrite file:", False],
            ["export_schema", "Export schema:", False],
        ]

    @property
    def args_strs(self):
        """Pass."""
        lines = []
        for arg, text, default in self.args_map:
            value = self.GETARGS.get(arg, default)
            if isinstance(value, str):
                value = repr(value)
            if isinstance(value, list):
                value = ", ".join(value)
                value = value or None
            lines.append(f"{text:30}{value}")
        return lines

    def __str__(self):
        """Show info for this object."""
        return f"{self.__class__.__name__} callbacks"

    def __repr__(self):
        """Show info for this object."""
        return self.__str__()



# -*- coding: utf-8 -*-
"""API model base classes and mixins."""
import abc
import math
import time

from ..constants import LOG_LEVEL_API, MAX_BODY_LEN, MAX_PAGE_SIZE
from ..exceptions import JsonError, JsonInvalid, NotFoundError, ResponseNotOk
from ..logs import get_obj_log
from ..tools import dt_now, dt_sec_ago, is_int, json_load, json_reload


class Model:
    """API model base class."""

    @abc.abstractproperty
    def router(self):
        """Router for this API model.

        Returns:
            :obj:`.routers.Router`: REST API route defs
        """
        raise NotImplementedError  # pragma: no cover


class ModelMixins(Model):
    """Mixins for :obj:`Model` objects."""

    def __init__(self, auth, **kwargs):
        """Mixins for :obj:`Model` objects.

        Args:
            auth (:obj:`.auth.Model`): object to use for auth and sending API requests
            **kwargs: passed to :meth:`Mixins._init`
        """
        log_level = kwargs.get("log_level", LOG_LEVEL_API)
        self.LOG = get_obj_log(obj=self, level=log_level)
        """:obj:`logging.Logger`: Logger for this object."""

        self.auth = auth
        """:obj:`.auth.Model`: object to use for auth and sending API requests."""
        self.http = auth.http

        self._init(auth=auth, **kwargs)

        auth.check_login()

    def _init(self, auth, **kwargs):
        """Post init method for subclasses to use for extra setup.

        Args:
            auth (:obj:`.auth.Model`): object to use for auth and sending API requests
        """
        pass

    def __str__(self):
        """Show info for this model object."""
        cls = self.__class__
        auth = self.auth.__class__.__name__
        url = self.http.url
        return f"{cls.__module__}.{cls.__name__}(auth={auth!r}, url={url!r})"

    def __repr__(self):
        """Show info for this model object."""
        return self.__str__()

    def _build_err_msg(self, response, error=None, exc=None):
        """Pass."""
        request_size = len(response.request.body or "")
        response_size = len(response.text or "")
        msgs = []
        msgs += [f"Original exception: {exc}"] if exc else []
        msgs += [
            f"Request Body:",
            json_reload(obj=response.request.body, error=False, trim=MAX_BODY_LEN),
            "",
            f"Response details:",
            f"  code: {response.status_code!r}",
            f"  reason: {response.reason!r}",
            f"  method={response.request.method!r}",
            f"  url: {response.url!r}",
            f"  request_size: {request_size}",
            f"  response_size: {response_size}",
        ]

        response_obj = json_load(obj=response.text, error=False)

        if isinstance(response_obj, dict):
            if "additional_data" in response_obj:
                msg = json_reload(obj=response_obj.pop("additional_data"), error=False)
                msgs += [f"  ** Additional Data:", msg]

            if "status" in response_obj:
                msgs += [f"  ** Status: " + response_obj.pop("status")]

            if "message" in response_obj:
                msgs += ["  ** Message: " + response_obj.pop("message")]

            msgs += [f"Extra: {response_obj}"] if response_obj else []
        else:
            msgs += [f"Response Body:", response_obj]

        msgs += [error] if error else []

        return "\n" + "\n".join(msgs)

    def request(
        self,
        path,
        method="get",
        raw=False,
        is_json=True,
        error_status=True,
        error_json_bad_status=True,
        error_json_invalid=True,
        # fmt: off
        **kwargs
        # fmt: on
    ):
        """Send a REST API request using :attr:`.auth.Mixins.http`.

        Args:
            path (:obj:`str`): path to use in request
            method (:obj:`str`, optional): default ``get`` - method to use in request
            raw (:obj:`bool`, optional): default ``False`` -

                * if ``True`` return the raw :obj:`requests.Response` object
                * if ``False`` return the text or json of the response based on is_json
            is_json (:obj:`bool`, optional): default ``True`` - if raw is False:

                * if ``True`` return the decoded json of the response text body
                * if ``False`` return the text body of the response
            error_status (:obj:`bool`, optional): default ``True`` -

                * if ``True`` check response status code
                  with :meth:`_check_response_code`
                * if ``False`` do not check response status code
            **kwargs:
                Passed to :meth:`.http.Http.__call__`

        Returns:
            :obj:`requests.Response` or :obj:`object` or :obj:`str`:

                * :obj:`requests.Response`: if raw is True
                * :obj:`object`: if raw is False and is_json is True
                * :obj:`str`: if raw is False and is_json is False
        """
        sargs = {}
        sargs.update(kwargs)
        sargs.update({"path": path, "method": method})

        response = self.http(**sargs)

        if raw:
            return response

        if is_json and response.text:
            data = self._check_response_json(
                response=response,
                error_json_bad_status=error_json_bad_status,
                error_json_invalid=error_json_invalid,
            )
        else:
            data = response.text

        self._check_response_code(response=response, error_status=error_status)

        return data

    def _check_response_code(self, response, error_status=True):
        """Check the status code of a response.

        Args:
            response (:obj:`requests.Response`): response object to check
            error_status (:obj:`bool`, optional): default ``True`` -

                * if ``True`` throw exc if response status code is bad
                * if ``False`` silently ignore bad response status codes

        Raises:
            :exc:`.ResponseNotOk`:
                if response has a status code that is an error and error_status is True
        """
        if error_status:
            try:
                response.raise_for_status()
            except Exception as exc:
                respexc = ResponseNotOk(
                    self._build_err_msg(
                        response=response,
                        error="Response has a bad status code!",
                        exc=exc,
                    )
                )
                respexc.exc = exc
                raise respexc

    def _check_response_json(
        self, response, error_json_bad_status=True, error_json_invalid=True
    ):
        """Check the text body of a response is JSON.

        Args:
            response (:obj:`requests.Response`): response object to check
            error_json_bad_status (:obj:`bool`, optional): default ``True`` -

                * if ``True`` throw an exc if response is a json dict that
                  has a non-empty error key or a status key that == error
                * if ``False`` ignore error and status keys in response json dicts
            error_json_invalid (:obj:`bool`, optional): default ``True`` -

                * if ``True`` throw an exc if response is invalid json
                * if ``False`` return the text of response if response is invalid json

        Raises:
            :exc:`.JsonInvalid`: if error_json_invalid is True and
                response has invalid json

            :exc:`.JsonError`: if error_json_bad_status is True and
                response is a json dict that has a non-empty error key or a
                status key that == error

        Returns:
            :obj:`object` or :obj:`str`:

                * :obj:`object` if response has json data
                * :obj:`str` if response has invalid json data
        """
        try:
            data = response.json()
        except Exception as exc:
            if error_json_invalid:
                respexc = JsonInvalid(
                    self._build_err_msg(
                        response=response, error="JSON is not valid in response", exc=exc
                    )
                )
                respexc.exc = exc
                raise respexc

            return response.text

        if isinstance(data, dict):
            has_error = data.get("error")
            has_error_status = data.get("status") == "error"

            if (has_error or has_error_status) and error_json_bad_status:
                respexc = JsonError(self._build_err_msg(response=response))
                raise respexc
        return data


class PagingMixins:
    """Pass."""

    def get(self, generator=False, **kwargs):
        """Get objects for a given query using paging.

        Args:
            generator (:obj:`bool`, optional): default ``False`` -

                * True: return an iterator for assets that will yield rows
                  as they are fetched
                * False: return a list of rows after all have been fetched
            **kwargs: passed to :meth:`get_generator`

        Yields:
            :obj:`dict`: row if generator is True

        Returns:
            :obj:`list` of :obj:`dict`: rows if generator is False
        """
        gen = self.get_generator(**kwargs)

        if generator:
            return gen

        return list(gen)

    def get_generator(self, **kwargs):
        """Pass."""
        raise NotImplementedError  # pragma: no cover

    def _get_page(self, state, store):
        page_start = dt_now()
        state["page_num"] += 1

        state["page"] = self._get(
            query=store["query"],
            page_size=state["page_size"],
            row_start=state["rows_fetched"],
        )
        rows = state["page"].pop("assets", [])

        self._get_page_handle(state=state, store=store, rows=rows)

        time_page = dt_sec_ago(obj=page_start)
        rows_page = len(rows)

        state["rows_start"] += rows_page
        state["rows_fetched"] += rows_page
        state["rows_page"] = rows_page
        state["time_page"] = time_page
        state["time_fetch"] += time_page

        self.LOG.debug("FETCHED PAGE: {}".format(state))

        return rows

    def _get_page_handle(self, rows, state, store):
        rows_total_page = state["page"].get("page", {}).get("totalResources")
        rows_total = state["rows_total"]
        page_size = state["page_size"]
        page_num = state["page_num"]

        if is_int(rows_total_page):
            rows_total = rows_total_page
        else:
            rows_total_page = rows_total

        page_total = math.ceil(rows_total_page / page_size)

        state["page_total"] = page_total
        state["page_left"] = page_total - page_num
        state["rows_total"] = rows_total

    def _get_page_size(self, page_size=MAX_PAGE_SIZE, max_rows=None):
        if max_rows and max_rows < page_size:
            self.LOG.debug(f"CHANGED PAGE SIZE {page_size} to max_rows {max_rows}")
            page_size = max_rows

        if page_size > MAX_PAGE_SIZE:
            self.LOG.debug(f"CHANGED PAGE SIZE {page_size} to max {MAX_PAGE_SIZE}")
            page_size = MAX_PAGE_SIZE

        return page_size

    def _get_page_stop_rows(self, row, rows, state, store):
        if state.get("stop", False):
            stop = state.get("stop_msg", "")
            self.LOG.debug(f"STOPPED ROW LOOP: STOP called {stop}")
            return True
        if state["max_rows"] and state["rows_processed"] >= state["max_rows"]:
            self.LOG.debug("STOPPED ROW LOOP: HIT MAX_ROWS")
            return True
        return False

    def _get_page_stop_fetch(self, rows, state, store):
        if not rows:
            self.LOG.debug(f"STOPPED FETCH LOOP: HIT NO ROWS state={state}")
            return True

        if state.get("stop", False):
            stop = state.get("stop_msg", "")
            self.LOG.debug(f"STOPPED FETCH LOOP: STOP called {stop} state={state}")
            return True

        if state["max_pages"] and state["page_num"] >= state["max_pages"]:
            self.LOG.debug(f"STOPPED FETCH LOOP: HIT MAX_PAGES state={state}")
            return True

        if state["max_rows"] and state["rows_fetched"] >= state["max_rows"]:
            self.LOG.debug(f"STOPPED FETCH LOOP: HIT MAX_ROWS state={state}")
            return True

        return False


class PagingMixinsObject(PagingMixins):
    """Pass."""

    def get_by_uuid(self, value, **kwargs):
        """Get a single saved query by name.

        Args:
            name (:obj:`str`): name of saved query to get
            **kwargs: passed to :meth:`get`

        Returns:
            :obj:`dict`: saved query
        """
        rows = self.get(**kwargs)

        for row in rows:
            if row["uuid"] == value:
                return row

        tmpl = "name: {name!r}, uuid: {uuid!r}".format
        valid = "\n  " + "\n  ".join(sorted([tmpl(**row) for row in rows]))
        raise NotFoundError(f"uuid {value!r} not found, valid:{valid}")

    def get_by_name(self, value, **kwargs):
        """Get a single saved query by name.

        Args:
            name (:obj:`str`): name of saved query to get
            **kwargs: passed to :meth:`get`

        Returns:
            :obj:`dict`: saved query
        """
        rows = self.get(**kwargs)

        for row in rows:
            if row["name"] == value:
                return row

        tmpl = "name: {name!r}".format
        valid = "\n  " + "\n  ".join(sorted([tmpl(**row) for row in rows]))
        raise NotFoundError(f"name {value!r} not found, valid:{valid}")

    def get_generator(
        self,
        query=None,
        max_rows=None,
        max_pages=None,
        page_size=MAX_PAGE_SIZE,
        page_start=0,
        page_sleep=0,
        **kwargs,
    ):
        """Get saved queries using paging.

        Args:
            query (:obj:`str`, optional): default ``None`` - filter rows to return

                This is NOT a query built by the query wizard!
            page_size (:obj:`int`, optional): default ``0`` - for paging, return N rows
            max_rows (:obj:`int`, optional): default ``None`` - return N assets

        Returns:
            :obj:`list` of :obj:`dict`: list of saved query metadata
        """
        page_size = self._get_page_size(page_size=page_size, max_rows=max_rows)

        store = {"query": query}
        state = {
            "max_pages": max_pages,
            "max_rows": max_rows,
            "page": {},
            "page_cursor": None,
            "page_left": 0,
            "page_num": page_start,
            "page_size": page_size,
            "page_sleep": page_sleep,
            "page_total": 0,
            "rows_fetched": 0,
            "rows_page": 0,
            "rows_processed": 0,
            "rows_start": page_start * page_size,
            "rows_total": 0,
            "time_fetch": 0,
            "time_page": 0,
        }

        self.LOG.info(f"START FETCH store={store}")

        while True:
            rows = self._get_page(state=state, store=store)
            for row in rows:
                yield row

                state["rows_processed"] += 1

                if self._get_page_stop_rows(
                    row=row, state=state, store=store, rows=rows
                ):
                    break

            if self._get_page_stop_fetch(rows=rows, state=state, store=store):
                break

            time.sleep(state["page_sleep"])

        self.LOG.info(f"FINISH FETCH store={store}")


class ChildMixins:
    """Mixins model for children of :obj:`Mixins`."""

    def __init__(self, parent):
        """Mixins model for children of :obj:`Model`.

        Args:
            parent (:obj:`Model`): parent API model of this child
        """
        self.parent = parent
        self.http = parent.http
        self.auth = parent.auth
        self.router = parent.router
        self.request = parent.request
        self.LOG = parent.LOG.getChild(self.__class__.__name__)
        self._init(parent=parent)

    def _init(self, parent):
        """Post init method for subclasses to use for extra setup.

        Args:
            parent (:obj:`Model`): parent API model of this child
        """
        pass

    def __str__(self):
        """Show info for this model object."""
        return f"{self.__class__.__name__} for {self.parent}"

    def __repr__(self):
        """Show info for this model object."""
        return self.__str__()



# -*- coding: utf-8 -*-
"""API models for working with adapters and connections."""
from ...constants import GENERIC_NAME
from ...tools import strip_right
from .config import parse_schema


def parse_adapters(raw):
    """Parser to turn adapters metadata into a more friendly format."""
    parsed = []

    for name, raws in raw.items():
        for raw in raws:
            adapter = parse_adapter(name=name, raw=raw)
            parsed.append(adapter)

    return parsed


def parse_adapter(name, raw):
    """Parse a single adapter."""
    parsed = {
        "name": strip_right(obj=name, fix="_adapter"),
        "name_raw": name,
        "name_plugin": raw["unique_plugin_name"],
        "node_name": raw["node_name"],
        "node_id": raw["node_id"],
        "status": raw["status"],
        "features": raw["supported_features"],
    }

    generic_name = GENERIC_NAME
    specific_name = get_specific_name(raw=raw)
    config = raw["config"]

    specific_schema = config.get(specific_name, {}).get("schema", {})
    specific_schema = parse_schema(raw=specific_schema)

    generic_schema = config[generic_name]["schema"]
    generic_schema = parse_schema(raw=generic_schema)

    cnx_schema = parse_schema(raw=raw["schema"])
    cnx_schema["connection_label"] = {
        "name": "connection_label",
        "title": "Connection Label",
        "type": "string",
        "required": False,
    }

    parsed["schemas"] = {
        "cnx": cnx_schema,
        "specific": specific_schema,
        "generic": generic_schema,
        "generic_name": generic_name,
        "specific_name": specific_name,
    }

    parsed["config"] = {
        "specific": raw["config"].get(specific_name, {}).get("config", {}),
        "generic": raw["config"].get(generic_name, {}).get("config", {}),
    }

    parsed["cnx"] = parse_cnx(raw=raw, parsed=parsed)
    parsed["cnx_count_total"] = len(parsed["cnx"])
    parsed["cnx_count_broken"] = len([x for x in parsed["cnx"] if not x["working"]])
    parsed["cnx_count_working"] = len([x for x in parsed["cnx"] if x["working"]])

    return parsed


def get_specific_name(raw):
    """Pass."""
    found = [x for x in raw["config"] if x != GENERIC_NAME]
    return found[0] if found else ""


def parse_cnx(raw, parsed):
    """Parse the connection metadata for this adapter."""
    cnx = []

    for cnx_raw in raw["clients"]:
        cnx_parsed = {}
        cnx_parsed["config"] = cnx_raw["client_config"]
        cnx_parsed["adapter_name"] = parsed["name"]
        cnx_parsed["adapter_name_raw"] = parsed["name_raw"]
        cnx_parsed["node_name"] = parsed["node_name"]
        cnx_parsed["node_id"] = parsed["node_id"]
        cnx_parsed["status"] = cnx_raw["status"]
        cnx_parsed["working"] = cnx_raw["status"] == "success" and not cnx_raw["error"]
        cnx_parsed["id"] = cnx_raw["client_id"]
        cnx_parsed["uuid"] = cnx_raw["uuid"]
        cnx_parsed["date_fetched"] = cnx_raw["date_fetched"]
        cnx_parsed["error"] = cnx_raw["error"]
        cnx.append(cnx_parsed)

    return cnx



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import click

from ..context import AliasedGroup
from . import (grp_central_core, grp_discover, grp_meta, grp_nodes, grp_roles,
               grp_settings, grp_users)


@click.group(cls=AliasedGroup)
def system():
    """Group: System control commands."""


system.add_command(grp_meta.meta)
system.add_command(grp_nodes.nodes)
system.add_command(grp_central_core.central_core)
system.add_command(grp_roles.roles)
system.add_command(grp_settings.settings_lifecycle)
system.add_command(grp_settings.settings_gui)
system.add_command(grp_settings.settings_core)
system.add_command(grp_users.users)
system.add_command(grp_discover.discover)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ....tools import json_dump
from ...context import CONTEXT_SETTINGS, click
from ...options import AUTH, add_options

ENABLED = click.option(
    "--enable/--no-enable",
    "-e/-ne",
    "enabled",
    help="Enable / disable the central core feature",
    default=None,
    is_flag=True,
    show_envvar=True,
    show_default=True,
    required=True,
)
DELETE_BACKUPS = click.option(
    "--delete-backups/--no-delete-backups",
    "-db/-ndb",
    "delete_backups",
    help="Delete backups once they are restored",
    default=None,
    is_flag=True,
    show_envvar=True,
    show_default=True,
    required=True,
)

OPTIONS = [
    *AUTH,
    ENABLED,
    DELETE_BACKUPS,
]


@click.command(name="update", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, **kwargs):
    """Update the central core configuration."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        data = client.system.central_core.update(**kwargs)

    click.secho(json_dump(data))
